#include <buffer.h>
#include <packetforwarding.h>
#include <bufferqueue.h>
#include <macros.h>
#include <simulation.h>
#include <utility/xmlattribute.h>
#include <sourcebuffer.h>
#include <triggernode.h>

#include <configuration.h>

#include <registercontainer.h>
#include <flit.h>
#include <packet.h>

#include <cassert>

REGISTER_ELEMENT(Buffer, "buffer")


bool Buffer::init(const DOMElement* scope)
{	
	bool fail = false;
	
	if(scope) setScope(scope);
	
	assert(Scope());

	if(!BQ)
	{
		BQ = new BufferQueue();
	}
	
	int min = getAttribute("minSpace");
	int max = getAttribute("maxSpace");
	if(max > min)
		fail = !BQ->resize(max);
	else
		fail = !BQ->resize(min);

	if(fail)
		_ERROR("Failed to set up buffer queue (" << toString() << ")!");
	
	return true;
}


double Buffer::Measure(TypeInfo type)
{
	switch (type) 
	{
	// Zeit bis Weitervermittlung (Paket)
	case eDelay:
	case ePrioriDelay:
		assert(_delay.valid());
		return _delay.value;
		break;
		
	// Zeit in Puffer (Paket)
	case eLatency:
	case ePrioriLatency:
		assert(_latency.valid());
		return _latency.value;
		break;
		
	// Laenge der Queue (Flits)
	case eQueue:
		return (double)BQ->spaceInUse();
		break;
	default:
		_ERROR("Unexpected measure type: " << type);
	}
	return 0.0;
}


void Buffer::takeMeasure(const Description& descr)
{	
	switch (descr.type) 
	{
	// Delay von Paketen mit angegebener Prioritaet
	case ePrioriDelay:
		if (_delay.priority != descr.priority) break;
				
	// Delay von Paketen in Puffer
	case eDelay:
		if (_delay.valid())
			Analysable::takeMeasure(descr);
		break;
		
	// Latenz von Paketen mit angegebener Prioritaet
	case ePrioriLatency:
		if (_latency.priority != descr.priority) break;
		
	// Latenz von Paketen in Puffer
	case eLatency:
		if (_latency.valid())
			Analysable::takeMeasure(descr);
		break;

	// Laenge der Queue
	case eQueue:
		Analysable::takeMeasure(descr);
	
	default:
		break;
	}
}

const Bitmask& Buffer::generateBitmask()
{	
	_bitmask = Out()[0]->To()->generateBitmask();

	return _bitmask;
}

void Buffer::reset()
{	
	BQ->clear();
}

int Buffer::putPacketOnLine(Wire* outwire, bool& head)
{	
	assert(rtti()!=Component::TARGET);
	
	// FIXME: Das muss einmal bei Initialisierung eines Puffers erfolgen
	if (!outwire->RegisterOnLine())
	{
		// Register auf die Line
		outwire->connectRegister(BQ->Register());
	}
	
	// Test, ob Flit zum anlegen vorhanden und bereit, oder ob keine Heads
	// von Paketen mehr gesendet werden drfen.
	if (BQ->empty())
	{
		return NOPACKET;
	}
	
	// Annahme: Zu versendendes Flit vorhanden. 
	assert(BQ->front() != NULL);
	// Annahme: Koroutine sorgt dafuer, dass Flits nur einmal
	// pro Takt bewegt werden.
	assert(BQ->front()->Location().since < (signed int)Tick());
	
	// ggf. Test, ob Deadline des zugehoerigen Paketes abgelaufen
	/*if (BQ->front()->Owner()->deadline < Tick())
	{
		_ERROR("Not re-implemented yet!", 111);

		firstPacket = BQ->pop(); // Register ist empty now
		assert(!BQ->registerSize());
     		
		delete firstPacket;
                S.deletedPacketCounter++;
                        
		return NOPACKET;
	}*/

	// Flag fuer Start eines Paketes
	head = BQ->front()->head();
	
	_PACKETMOVE("Offered flit:", outwire->getFlit()->toString(),
			"From: " << toString());
			
	return TriggerNode::SUCCESS;
}

int Buffer::takePacketFromLine(Wire* inwire, bool& tail)
{	
	// angelegtes und abgeholtes Flit
	const Flit* fetchable = NULL;
	Flit* fetched = NULL;

	// Register auf Eingangsleitung nach Flit fragen
	fetchable = inwire->getFlit(this);
	
	// Test, ob Flit fuer diesen Knoten verfuegbar
	if (!fetchable)
	{
		return TriggerNode::NOPACKET;
	}
		
	// Test, ob Flit bereit
	// Mgliches Abweisen des Pakets im PREPARE-Status, wenn dynamisch
	// (Einzige Stelle, wo dynamisches/statisches Rekonfigurieren unterschiedlich 
	// bezglich der Paketbewegung. Analyse und Paketkopieren in simulation.cpp stellen
	// keine eigentlichen Paketbewegungen dar.)
	if(!fetchable->moveable() || (fetchable->head() && State()==Simulation::PREPARE 
				&& Simulation::Instance()->Active()->Dynamic()))
	{	
		// Annahme: Wenn Flit nicht bereit, dann muss ein Puffer-Knoten
		// Vorgaenger sein, da Router nur auf einen Ausgang schaltet, wenn 
		// Flit auch bewegt werden kann.
		//assert(inwire->From()->rtti() <= Component::BUFFER);
		
		// Annahme: Register auf Leitung
		assert(inwire->RegisterOnLine());

		// Flit kann nicht aufgenommen werden
		inwire->RegisterOnLine()->rejectFlit(this);
		
		_PACKETMOVE("Unmoveable flit:",inwire->getFlit(this)->toString(),
				"From: " << inwire->From()->toString() << std::endl
					<< " To: " << toString());
		
		return TriggerNode::DISMISS;
	}

	// Test, ob Flit aufgenommen werden kann	
	if (!BQ->freeSpace() || (BQ->estimatedSpace() < fetchable->neededSpace())) 
	{
		// FIXME: Nur Pakete passender Groesse generieren!
		if(fetchable->neededSpace() > BQ->queueSpace())
			_ERROR("Packet to large for buffering!" << std::endl 
 	  	        	<< "\t-> " << fetchable->toString() << std::endl 
				<< "\t-> " << toString() << std::endl 
				<< "\tHINT: Increase buffer space by setting attribute 'maxSpace'.");
		
		_PACKETMOVE("Rejected flit:",inwire->getFlit(this)->toString(),
				"From: " << inwire->From()->toString() << std::endl
				<< " To: " << toString());
		
		assert(inwire->RegisterOnLine());

		// Flit kann nicht aufgenommen werden
		inwire->RegisterOnLine()->rejectFlit(this);
		
		return TriggerNode::DISMISS;
	}

#ifdef PACKETMOVE
	std::stringstream from;
	std::stringstream to;
	std::stringstream info;
	Buffer* dummy = fetchable->Location().where;
 
	info << fetchable->toString();
	from << "From: " << dummy->toString() << " (" << dummy->BQ->freeSpace() << "/";
	to << " To: " << toString() << " (" << BQ->freeSpace() << "/";
#endif
	
	fetched = inwire->RegisterOnLine()->fetchFlit(this);
	
	tail = fetched->tail();
	
	flitReceived(fetched);
	
	
#ifdef PACKETMOVE
	from << dummy->BQ->freeSpace() << ")";
	to << BQ->freeSpace() << ")";

	_PACKETMOVE("Moved flit:", info.str(), from.str() << std::endl << to.str());
#endif
	return TriggerNode::SUCCESS;
}

int Buffer::lineActivated(Wire* inwire, ModifierList*, SignedTNodeList* list, TNodeList::iterator pos)
{
	// Annahme: Leitung ist eingehende Leitung dieses Puffers
	assert(inwire->getInIndex() < In().size() && In()[inwire->getInIndex()] == inwire);
	
	assert(inwire->RegisterOnLine());

	if(inwire->getPacket())
		inwire->RegisterOnLine()->registerCustomer(this, new Bitmask(_bitmask));
		
	// moeglicher Empfaenger wird verlangt
	if(list && inwire->getFlit(this))
		list->insert(pos,this);
		
	// Nachricht, dass wir Signal erhalten haben
	return TriggerNode::SUCCESS;
}

int Buffer::lineFinished(Wire* cleared, ModifierList*, 
				SignedTNodeList* list, TNodeList::iterator pos)
{
	if(list && pos !=NLITNULL && BQ->Register()->finalState())
	{
		list->insert(pos, this);
	}
	
	return TriggerNode::SUCCESS;
}

void Buffer::flitReceived(Flit *flit)
{
	// Flit aktualisieren..
	flit->buffered(this, Tick());
	// ..und einpuffern
	BQ->push(flit);
}

void Buffer::flitTransmitted(Flit *flit, bool complete)
{
	if(!complete) return;
	
	assert(flit && flit==BQ->front());
	
	// Messungen fuer Latenz und Verzoegerung aktualisieren
	unsigned int priority = flit->Owner()->priority;
 	
	if(flit->head())
	{
		_delay.priority = priority;
		_delay.value = Tick() - flit->Location().since;
		_delay.sign();
	}
	if(flit->tail())
	{
		_latency.priority = priority; 
		_latency.value = Tick() - flit->Owner()->at(0)->Location().since;
		_latency.sign();
	}
	
	// Flit entfernen
	BQ->pop();
}

