#include <gbforwarding.h>
#include <configuration.h>
#include <targetbuffer.h>
#include <bufferqueue.h>
#include <sourcebuffer.h>
#include <wirematrixiterator.h>
#include <wirematrixentry.h>
#include <macros.h>
#include <utility/xmlhelper.h>
#include <utility/random.h>
#include <routing.h>
#include <switching.h>
#include <simulation.h>
#include <flit.h>

#include <utility/bitmask.h>
#include <triggernode.h>

#include <stack>
#include <assert.h>

REGISTER_ELEMENT(GBForwarding, "GlobalBackpressure")

// Aktivierungsbaeme aufbauen, ForwardingModifier generieren
ForwardingModifier* GBForwarding::initForwarding()
{
	// Aktivierungsbaeme aufbauen
	initActivationTrees();

	return new GBForwardingModifier;
}


void GBForwarding::start()
{
	receivers.clear();
	
	// Bevor jegliche PF-Planung oder Bewegung gemacht wurde, sind alle Puffer
	// ohne Vormerkung jeglicher Umorientierung.
	checked.clear();
	ComponentVector buffers = Simulation::Instance()->Active()->Components(Component::BUFFER);
	for(unsigned int n = 0; n<buffers.size(); n++)
		checked.push_back(false);
	
	// Initial knnen nur Targets empfangen, dazu in zuflliger 
	// Reihenfolge zur Liste receivers hinzufuegen
	// Bei dynamischer Rekonfiguration muss die Reihenfolge regelmssig sein.
	if(Simulation::Instance()->Handler()->Critical())
	{
		for(unsigned int ini=0; ini<initial.size(); ini++)
		{
			TriggerNode* root = initial[ini];
			root->set(TriggerNode::ACTIVE);
			receivers.push_back(root);
		}
	}
	else
	{
    		RandomKofN r(initial.size());
		while(!r.Empty())
		{			
			// select a root element
			TriggerNode* root = initial[r.Choose()-1];
		
			root->set(TriggerNode::ACTIVE);
		
			receivers.push_back(root);
		}
	}
	receivePackets();
}


void GBForwarding::receivePackets()
{
	// empfangsbereiter Knoten und zugehoeriger Eingang
	TriggerNode* r_node;
	Wire* inwire = NULL;
	bool tailOnLine = false;
	ModifierList finish;
	int status = -1;
	
	// Solange empfangsbereite Knoten existieren
	while(!receivers.empty())
	{
		// erster Knoten aus Liste der moeglich Empfaenger
		r_node = receivers.front();
		receivers.pop_front();	
		
		// Annahme: Knoten hat gesendet aber nicht empfangen
		assert(r_node->test(TriggerNode::PULLED));
		assert(!r_node->test(TriggerNode::PUSHED));
				
		// Kein Eingang, kein Paket zu empfangen
		if(!r_node->In().size())
			continue;
		
		// Eingangsleitung
		inwire = r_node->In().front();

		// Vorgaenger ggf. aktivieren
		activateChildren(r_node,receivers.begin());
		
		// Annahme: Empfangsbereite Knoten haben noch nicht empfangen
		// in request(tnode): assert(!tnode->test(TriggerNode::PUSHED));
		switch(request(r_node))
		{
			// Paket liegt an
			case TriggerNode::SUCCESS:
			{
				TriggerNode* prev = r_node->In().front()->getFlit(r_node)->Location().where;

				const Bitmask bmP = r_node->In().front()->getFlit(r_node)->Owner()->getBitmask();
				
				// Empfangen nur wenn der Sender noch senden soll (SaF). 
				// Empfangen immer (pCT)
				if(!prev->test(TriggerNode::PULLED) || (prev->rtti()==1 && ((Buffer*)prev)->BQ->infinite) 
						|| isPCT || isVCT)
				{
					// Versuch zu empfangen
					status = r_node->takePacketFromLine(inwire, tailOnLine);
								
					// Koroutine sollte dafuer sorgen, dass Paket anliegt 
					// und Knoten dieses aufnehmen kann oder mangels Platz 
					// oder vergebener Ressourcen verwirft
					assert(status != TriggerNode::NOPACKET);
				
					// Tail wurde empfangen
					if(status == TriggerNode::SUCCESS && tailOnLine)
					{
						//Buffer und seinen Vorgnger zum Beschneiden vormerken.
						if(r_node->rtti()==1 && ((Buffer*)r_node)->BQ->infinite)
							((Buffer*)r_node)->BQ->cut.sign();
						if(prev->rtti()==1 && ((Buffer*)prev)->BQ->infinite)
							((Buffer*)prev)->BQ->cut.sign();

						// Buffer, von dem das Paket kam, zum Auflsen des Maeander-Flags vormerken.
						if(prev->rtti()==1 && ((Buffer*)prev)->maeander)
							((Buffer*)prev)->solve_maeander.sign();

						finish.push_back(switching()->clear());
					}
				}
			}
			
			// Kein Paket liegt an, Leitung freigeben
			case TriggerNode::NOPACKET:
			{			
				// Vorgaenger informieren, Liste empfangsbereiter 
				// Knoten fuer diesen Takt aktualisieren
				inwire->From()->lineFinished(inwire,&finish,
						&receivers,receivers.begin());
				
				// Vorschlag nderung 2: PUSHED muss hier nicht gesetzt werden.
				r_node->set(TriggerNode::PUSHED);

				// Ein PUSHED Buffer kann nicht das Ziel einer Bezugsnderung sein.
				if(r_node->rtti()==1)
					checked[r_node->Index()]=true;
				
				finish.clear();
				break;
			}
			
			// TODO: Noch keine Entscheidung
			//case TriggerNode::WAIT:

			default:
				_ERROR("Unexpected status after request for packet!");
		}
	}
	
#ifdef PACKETMOVE
	
	const TNodeVector& buffers = (TNodeVector&)Components(Component::BUFFER);
	const TNodeVector& sources = (TNodeVector&)Components(Component::SOURCE);
	const TNodeVector& targets = (TNodeVector&)Components(Component::TARGET);
	
	for(unsigned int i=0; i<buffers.size();i++)
	{
		if(buffers[i]->Out().size() && !buffers[i]->test(TriggerNode::ACTIVE))
			_ERROR(buffers[i]->toString() << " not activated in this tick!");

		if(buffers[i]->Out().size() && !buffers[i]->test(TriggerNode::PULLED))
			_ERROR(buffers[i]->toString() << " not sent in this tick!");

		if(buffers[i]->In().size() && !buffers[i]->test(TriggerNode::PUSHED))
			_ERROR(buffers[i]->toString() << " not received in this tick!");
	}

	for(unsigned int i=0; i<targets.size();i++)
	{
		if(!targets[i]->test(TriggerNode::ACTIVE))
			_ERROR(targets[i]->toString() << " not activated in this tick!");

		if(targets[i]->In().size() && !targets[i]->test(TriggerNode::PUSHED))
			_ERROR(targets[i]->toString() << " not received in this tick!");
	}
	
	for(unsigned int i=0; i<sources.size();i++)
	{
		if(sources[i]->Out().size() && !sources[i]->test(TriggerNode::ACTIVE))
			_ERROR(sources[i]->toString() << " not activated in this tick!");
		
		if(sources[i]->Out().size() && !sources[i]->test(TriggerNode::PULLED))
			_ERROR(sources[i]->toString() << " not sent in this tick!");

	}
#endif

}

void GBForwarding::activateChildren(TriggerNode* parent, TNodeList::iterator pos)
{
	ModifierList none;
	TNodeList resolved;
	
	const TNodeVector& tmp = parent->children();
	for(unsigned int i=0; i<tmp.size(); i++)
	{
		TriggerNode* a_node = tmp[i];
				
		if(a_node->test(TriggerNode::ACTIVE))
			continue;
					
		// Einschrnkung fr nicht-kritische Netze zur optimierung
		// der Laufzeit und Paketbewegung
		
		switch(activate(a_node))
		{
			case TriggerNode::DISMISS:
			case TriggerNode::NOPACKET:
			{
				// Kein Paket zu vermitteln, also kann empfangen werden
				receivers.insert(pos, a_node);		
				activateChildren(a_node, pos);
				break;
			}
				
			case TriggerNode::SUCCESS:
			{
				// Paket vorhanden und auf die Line gelegt,
				// warten bis es vom Nachfolger empfangen oder 
				// verworfen wird.
				
				// MaeanderPF-Methode: Suche neue Bezugspuffer.
				
				if(a_node->rtti()==1 && ((Buffer*)a_node)->maeander && !checked[a_node->Index()])
				{
					vector<Buffer*> optRxBuf = spaceSuccessors((Buffer*)a_node);
					if(optRxBuf.size())
					{
						TriggerNode* tmp2 = NULL;
						for(unsigned int jj=0;jj<optRxBuf.size();jj++)
						{
							tmp2 = (TriggerNode*)optRxBuf[jj]; 
							if(!tmp2->test(TriggerNode::PULLED))
								break;
						}
								
						if(!tmp2->test(TriggerNode::ACTIVE))
							activate(tmp2);
						
						receivers.insert(receivers.begin(), tmp2);
					}
				}
				
				break;
			}
			default:
				_ERROR("Unexpected return from method activate()!");
		}
		
	}
}

vector<Buffer*> GBForwarding::spaceSuccessors(Buffer* source)
{
	vector<Buffer*> rtn;
	assert(!source->BQ->empty()); // Leere Puffer drfen nicht gefragt werden!
	
	// Targets oder schon geprfte Buffer returnieren mit Antwort "{}".
	if(source->rtti()!=1 || checked[source->Index()]) return rtn;
	
	unsigned int sourceFrontNr = source->BQ->front()->Owner()->number;
	bool nothead = !source->BQ->front()->head();
	
	// Sobald ich den Puffer prfe, setze ich die Marke.
	checked[source->Index()] = true;
	
	// Nimm alle Nachfolge-Puffer.
	Node::Filter* nf = new Buffer::Filter();
	NodeVector nv = source->nextNodes(nf);

	// Wenn es keine gibt, dann Antwort "{}".
	if(nv.empty()) return rtn;
	
	// Wenn source kein head, dann iterativ nichtvollen Nachfolger suchen.
	// Verwende hier freeSpace(), da entweder tatschlicher freier Platz vorhanden oder
	// infinite-Puffer solchen "vorgaukeln".
	if(nothead)
	{	
		/* Suche den Head des Pakets */
		bool seek_error = false;
		bool found_stop = false;
		Buffer* finger = NULL;
		unsigned int fingerFrontNr = sourceFrontNr;
		// solange der Head nicht ausfindig gemacht, kein infinite-Puffer angetroffen 
		// und kein Suchfehler aufgetreten ist...
		while(nothead && !seek_error && !found_stop)
		{
			finger = NULL;
			// Durchsuche die Nachfolgepuffer
			for(unsigned int kk=0; kk<nv.size(); kk++)
			{
				Buffer* nextBuf = (Buffer*)(nv[kk]);
				// Gefunden, wenn einer die gleiche Back-Nummer wie die source hat.
				if(!nextBuf->BQ->empty() && fingerFrontNr == nextBuf->BQ->back()->Owner()->number && nextBuf!=source)
				{
					finger = nextBuf;
					break;
				}
			}
			// wenn gefunden, update die benutzten Variablen, die auf finger bezogen sind.
			if(finger)
			{
				// nothead, wenn die Front kein Head ist.
				nothead = !finger->BQ->front()->head();
				fingerFrontNr = finger->BQ->front()->Owner()->number;
				nv = finger->nextNodes(nf);
				found_stop = finger->BQ->infinite;
				// Suchfehler, wenn geprften Buffer erreicht.
				seek_error = seek_error || (finger->rtti()==1 && checked[finger->Index()]);
			}
			else seek_error = true;
		}
		
		// Ein Suchfehler hat zur Folge, dass ich mit {} antworte.
		if(seek_error) return rtn;
		
		rtn.push_back(finger);	
		return rtn;
	}
	
	const Bitmask* sourceBM = &source->BQ->front()->Owner()->getBitmask();

	// Andernfalls suche ungeprfte, die noch Platz haben und die nvBuf.I == source.I
	for(unsigned int kk=0; kk<nv.size(); kk++)
	{
		Buffer* nvBuf = (Buffer*)(nv[kk]);

		// Weg fr source-Paket ber nvBuf erlaubt (bzw. sinnvoll)?
		Bitmask* bValid = 
			Simulation::Instance()->Active()->Coroutine()->routing()->validOutput(
				NULL,
				sourceBM,
				nvBuf->In()[0]);
		
		if(!checked[nvBuf->Index()] && nvBuf->rtti()==1 && nvBuf->BQ->freeSpace() 
				&& nvBuf->BQ->infinite==source->BQ->infinite && bValid)
			rtn.push_back(nvBuf);
	}
	// Wenn gefunden, gib die Liste zurck.
	if(!rtn.empty()) return rtn;
	
	// Letzte Mglichkeit: Es gibt Nachfolgepuffer, doch sie sind alle voll
	// oder schon geprft. Frage nach spaceSuccessors der vollen, die mit "{}"
	// antworten. Antworten vereinigen und zurckgeben.
	for(unsigned int kk=0; kk<nv.size(); kk++)
	{
		Buffer* nvBuf = (Buffer*)(nv[kk]);
		
		// Weg fr source-Paket ber nvBuf erlaubt (bzw. sinnvoll)?
		Bitmask* bValid = 
			Simulation::Instance()->Active()->Coroutine()->routing()->validOutput(
				NULL,
				sourceBM,
				nvBuf->In()[0]);

		// Tiefensuche, wenn bValid, nvBuf.I == finger.I und nvBuf nicht checked
		if(nvBuf->BQ->infinite==source->BQ->infinite && !checked[nvBuf->Index()] && nvBuf->rtti()==1 && bValid)
		{
			vector<Buffer*> tmpVB = spaceSuccessors(nvBuf);
			if(tmpVB.size()==1)
				rtn.insert(rtn.end(),tmpVB[0]);
			else if(tmpVB.size()>1)
				rtn.insert(rtn.end(),tmpVB.begin(),tmpVB.end());
		}
	}
			
	return rtn;	
}

int GBForwarding::activate(TriggerNode* tnode)
{
	// Annahme: Zu aktivierende Knoten haben noch nicht gesendet
	assert(!tnode->test(TriggerNode::PULLED));
	
	ModifierList matrixinit;
	bool headOnLine = false;
	
	// Nur Targets haben keine Nachfolger, werden aber normalerweise
	// von Koroutine nicht ueber diese Methode aktiviert.
	if(tnode->Out().size() != 1)
		return TriggerNode::ERROR;

	// Der nachfolgende Netzknoten
	Node* next = tnode->Out().front()->To(); 
	
	// Paket auf die Line
	int state = tnode->putPacketOnLine(tnode->Out().front(), headOnLine);

	switch(state)
	{
		case TriggerNode::SUCCESS:
		{
			if(headOnLine)
				matrixinit.push_back(routing());
		}
		case TriggerNode::NOPACKET:break;
		default:
			_ERROR("Unexpected state after activating line!");
	}
	
	// Nachfolger benachrichtigen, ggf. Routing durchfuehren	
	next->lineActivated(tnode->Out().front(), &matrixinit);

	// Knoten aktivieren
	tnode->set(TriggerNode::ACTIVE);

	return state;
}

int GBForwarding::request(TriggerNode* tnode)
{
	// Annahme: Empfangsbereite Knoten haben noch nicht empfangen
	assert(!tnode->test(TriggerNode::PUSHED));
	
	// keine eingende Leitung, kein Paket
	if(!tnode->In().size())
	{
		return TriggerNode::NOPACKET;
	}
	
	// Der Vorgaenger
	Node* prev = tnode->In().front()->From();

	// kein Register auf Leitung, also keine Schaltung
	if(!tnode->In().front()->RegisterOnLine())
	{
		bool headOnLine = false;
		
		// angeschlossenen Knoten (Router) aktivieren
		int result = prev->putPacketOnLine(tnode->In().front(), headOnLine);
		
		// Annahme: Neue Schaltung nur fuer Head eines Paketes
		assert(result != TriggerNode::SUCCESS || headOnLine);    
		
		return result;
	}

	if(tnode->In().front()->getFlit(tnode))
		return TriggerNode::SUCCESS;

	return TriggerNode::NOPACKET;
}

void GBForwarding::initActivationTrees()
{
	std::stack<TriggerNode*> reached;
	const ComponentVector& targets = Components(Component::TARGET);
	
	// Targets wurdern erreicht...
	for(unsigned int i=0; i<targets.size(); i++)
	{
		reached.push((TriggerNode*)targets[i]);

		// Targets sind immer Ausgangspunkt fuer Aktivierung.
		initial.push_back((TriggerNode*)targets[i]);
	}
	
	// Nur TriggerNodes
	Node::Filter* filter = new TriggerNode::Filter;
	while(!reached.empty())
	{
		TriggerNode* r_node = reached.top();
		reached.pop();
		const NodeVector& tmp = r_node->prevNodes(filter);
		for(unsigned int i=0; i<tmp.size(); i++)
		{
			TriggerNode* r_node_child = (TriggerNode*)tmp[i];
			r_node->addChild(r_node_child);
			if(!r_node_child->hasChildren())
				reached.push(r_node_child);
		}
	}
	delete filter;
}

int GBForwardingModifier::precondition(WireMatrixIterator* it) const
{
	WireMatrixEntry& entry = it->value();
	
	if(entry.getOutWire()->To()->rtti() == Component::ROUTER)
		_ERROR("Can not handle a router at outgoing wire!");
	
	if(!((TriggerNode*)entry.getOutWire()->To())->test(TriggerNode::PULLED))
		return SKIPPED;
	
	return entry.test(I_PACKET) && entry.test(IO_VALID);
}

