#ifndef __LOCALBACKPRESSURE_H__
#define __LOCALBACKPRESSURE_H__

#include <packetforwarding.h>

//! Schaltmatrix-Manipulator fuer zulaessige Paketbewegung.
/*!
 * Diese Klasse sorgt fuer die Anpassung der Schaltmatrix eines Routers bezueglich 
 * der Bedingungen fuer das lokale Backpressure. Dazu wird das Bit IO_FIT eines 
 * Einganges gesetzt, wenn der Netzknoten am zugehoerigen Ausgang ein zu vermittelndes 
 * Paket des zueghoerigen Einganges aufnehmen kann. Diese Entscheidung kann fuer lokales
 * Backpressure immer getroffen werden, da die Pufferzustnde vom Ende des letzen Taktes
 * betrachtet werden.
 *
 * @see WireMatrixEntry
 * 
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class LBForwardingModifier : public ForwardingModifier
{
public:
	
	//! Re-Implementierung der abstrakten Basisklasse WireMatrixModifier.
	/*!
	 * Damit ueberhaupt geprueft werden kann, ob der nachfolgende Puffer 
	 * ein Paket aufnehmen kann, muss ein Paket vorliegen, dass den den 
	 * Ausgang zum uebergebenen Schaltmatrixeintrag benutzen will. 
	 * 
	 * return PASSED, falls Bits IO_VALID und I_PACKET gesetzt, FAILED sonst.
	 */
	int precondition(WireMatrixIterator*) const;
};


//! Koroutine fuer lokales Backpressure.
/*!
 * Diese Klasse implementiert eine Koroutine der Simulation, die fuer die Bewegung 
 * der Pakete im Netz waehrend eines Taktes nach der Strategie des lokalen
 * Backpressure sorgt. Dazu werden die Fuellstaende der Puffer im Netz betrachtet, 
 * wie sie am Ende des letzten Taktes vorlagen.
 *
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class LBForwarding : public PacketForwarding
{
public:
	
	//! Konstruktor.
	/*!
	 * Der Konstruktor bekommt als Argument einen Ausschnitt des eingelesenen 
	 * XML-Dokuments in Form eines Element-Knotens des vom Parser erstellten 
	 * DOM-Baumes. Dieser Knoten (oder untergeordnete Knoten) liefert 
	 * Informationen zu Routing und Switching der Simulation.
	 * 
	 * @param scope XML-Element mit Spezifikation der Koroutine.
	 */
	LBForwarding(const DOMElement* scope=NULL) 
		: PacketForwarding(scope) {}
		
	//! Implementierung der abstrakten Basisklasse PacketForwarding.
	/*!
	 * Zur Realisierung von lokalem Backpressure werden alle Puffer des
	 * Netzes aufgefordert, Pakete an ihre ausgehenden Leitungen anzulegen.
	 * Anschliessend werden die Puffer in zufaelliger Reihenfolge aufgefordert 
	 * Pakete anzufordern, wodurch die Router des Netzes schalten. Dadurch 
	 * liegen alle Pakete, die in diesem Takt versendet werden koennen, auf den 
	 * entsprechenden Leitungen und koennen von den empfangsbereiten Puffern 
	 * aufgenommen werden.
	 *
	 * @see PacketForwarding
	 */
	void start();
	
	
protected:

	/*!
	 * Implementierung der abstrakten Basisklasse PacketForwarding.
	 *
	 * @return Instanz der Klasse LBForwardingModifier.
	 */
	ForwardingModifier* initForwarding()
	{
		return new LBForwardingModifier;
	}
	
	//! Aktivieren eines Knotens.
	/*!
	 * Das Aktivieren eines Knotens sorgt dafuer, dass geprueft wird, ob
	 * ein Flit eines Paketes in diesem Takt zu versenden ist. Ist dies der Fall, 
	 * so wird dieses entsprechend an eine ausgehende Leitung angelegt, was
	 * dazu fuehrt, dass bei einem Router, im Falle eines Heads mit Routing-Informationen, 
	 * die Schaltmatrix entsprechend des verwendeten Routings aktualisiert wird.
	 *
	 * @param tnode Ein zu aktivierender Knoten.
	 * @return Status (e_trigger_return).
	 */
	int activate(TriggerNode* tnode);
	
	//! Anfordern eines Paketes.
	/*!
	 * Bevor ein Flit empfangen werden kann, signalisiert ein Knoten seine
	 * Bereitschaft seinem Vorgaenger. Das sorgt bei einem Router dazu, dass 
	 * um den entsprechenden Ausgang bei Bedarf ein Scheduling der relevanten 
	 * Eingaenge durchgefuehrt wird. Ist ein Flit verfuegbar und der
	 * verfuegbare Platz reicht aus, so wird der anfordernde Knoten in die
	 * Liste empfangsbereiter Knoten aufgenommen.
	 *
	 * @param tnode Ein anfordernder Knoten.
	 * @return Status (e_trigger_return).
	 */
	int request(TriggerNode* tnode);
	
	//! Empfangen der Pakete.
	/*!
	 * Das Empfangen von Paketen sorgt fuer die Paketbewegung im Netz. Dazu 
	 * wird eine Liste aller empfangsbereiten Knoten durchlaufen und die 
	 * anliegenden Pakete bewegt. Die Liste umfasst nur die tatschlich
	 * empfangsbereiten Knoten im Hinblick auf die Pufferzustaende zum Ende
	 * des letzten Taktes.
	 */
	void receivePackets();

private:
	
	//! Liste empfangsbereiter Knoten in einem Takt.
	SignedTNodeList receivers;
};

#endif

