#include <dlhandler.h>
#include <registercontainer.h>
#include <bufferqueue.h>
#include <packetforwarding.h>
#include <component.h>
#include <routing.h>
#include <utility/bitmask.h>

TreeNode::TreeNode(int givenNr, TreeNode* wherefrom)
{
	nr = givenNr;
	last = wherefrom;
	if(wherefrom!=0)
	{
		// Den Path weiss der node, wenn er gebaut wird,
		// die Liste der Bltter nicht. Letzteres wird 
		// erst beim Analysieren jedem node zugewiesen.
		path = wherefrom->path;
		path.push_back(wherefrom->nr);
	}
}

unsigned int TreeNode::passedFor(unsigned int testNr)
{
	for(unsigned int itmp=0;itmp<path.size();itmp++)
		if(path[itmp]==testNr)
			return path.size()-itmp;

	return 0;
}

downProtocol TreeNode::downAnalyze()
{
	downProtocol DP;
	
	// wenn der node ein Blatt ist, dann ist er FREE
	// und braucht this in seiner leaves Liste. 
	// Die einzigen, die sich selbst in ihrer Liste haben, ohne 
	// loops darzustellen, sind die Bltter!
	if(isLeaf()) 
	{
		DP.downType = FREE;
		DP.leaves.push_back(this);
		return DP;
	}
	
	// Der node ist eventuell eine Astgabel. D.h. wir mssen:
	// - die leaveNr Listen vereinigen
	// - Falls Nachfolger nur F oder N
	//    N, wo leaf=root: SOLVABLE
	//    andernfalls:     NORMAL
	// - andernfalls
	//    N, wo leaf=root: DIVISIBLE
	//    andernfalls: 	UNSOLVABLE
	//
	//		   | passable  | not passable |
	//   ~~~~~~~~~~~~~~|~~~~~~~~~~~|~~~~~~~~~~~~~~|
	//   not intricate | SOLVABLE  | NORMAL       |
	//   intricate     | DIVISIBLE | UNSOLVABLE   |
	//
	//   (Siehe dlhandler.h)
		
	bool intricate = false;
	bool passable = false;
	
	for(vector<TreeNode*>::iterator tnit=next.begin(); tnit!=next.end(); tnit++)
	{
		// Analysiere den i-ten Nachfolger
		downProtocol DPtmp = (*tnit)->downAnalyze();

		// Wenn ein Nachfolger nicht NORMAL oder FREE ist, dann ist this intricate.
		if(DPtmp.downType != NORMAL && DPtmp.downType != FREE)
			intricate = true;

		// Passable sind alle Bume, die mit der jetzigen Iteration einen Ring-Pfad
		// vom Root zu einem Leaf haben.
		
		// Wenn ein Nachfolger NORMAL oder UNSOLVABLE (d.h. not passable) und this noch
		// nicht als passable erkannt, dann finde heraus, ob nr \in DPtmp.leaves
		if((DPtmp.downType == NORMAL || DPtmp.downType == UNSOLVABLE)  && !passable)
			for(unsigned int ui = 0; ui<DPtmp.leaves.size(); ui++)
				if(DPtmp.leaves[ui]->nr == nr)
				{
					// Wenn ja, wird this zu passable
					passable = true;
					break;
				}
		// Anmerkung: Dies funktioniert nur, weil loops immer Bltter betreffen. Also
		// wenn ich loop mit root finde, dann weiss ich, dass kein innerer loop da ist.

		// erste Iteration: Die vorlufige Liste DP.leaves wird vom Nachfolger 0 bernommen.
		if(DP.leaves.size()==0)
		{
			DP.leaves = DPtmp.leaves;
		}
		else
		{
			// alle weiteren Iterationen: vereinige bisherige Liste DP.leaves mit der
			// Liste DPtmp.leaves des i-ten Nachfolgers. Die vereinigte Liste soll nun
			// immer in aufsteigender Reihenfolge (gemss nr) sortiert bleiben.
			// (allerdings: merke dir gleich grosse Bltter einzeln!)
			downProtocol DPjoin;
			unsigned int indDP = 0;
			unsigned int indDPtmp = 0;
			while(indDP < DP.leaves.size() && indDPtmp < DPtmp.leaves.size())
			{
				// wenn keine Liste schon ihr Ende erreichte
				// fge immer das kleinere Element hinzu
				if(DP.leaves[indDP]->nr < DPtmp.leaves[indDPtmp]->nr)
				{
					DPjoin.leaves.push_back(DP.leaves[indDP]);
					indDP++;
				}
				else // if DP.leaves[indDP]->nr >= DPtmp.leaves[indDPtmp]
				{
					DPjoin.leaves.push_back(DPtmp.leaves[indDPtmp]);
					// oder das "gleich grosse", und zwar beide Bltter!
					if(DP.leaves[indDP]->nr == DPtmp.leaves[indDPtmp]->nr)
					{
						DPjoin.leaves.push_back(DP.leaves[indDP]);
						indDP++;
					}
					indDPtmp++;
				}	
			}
			// wenn eine Liste schon fertig, fge den Rest der anderen hintendran.
			if(indDP < DP.leaves.size())
			{
				while(indDP < DP.leaves.size())
				{
					DPjoin.leaves.push_back(DP.leaves[indDP]);
					indDP++;
				}
			}
			else
			{
				while(indDPtmp < DPtmp.leaves.size())
				{
					DPjoin.leaves.push_back(DPtmp.leaves[indDPtmp]);
					indDPtmp++;
				}
			}
			DP.leaves = DPjoin.leaves;		
		}	
	}
	
	// Welcher Typ ist es?
	if(intricate)
		if(passable)
			DP.downType = DIVISIBLE;
		else
			DP.downType = UNSOLVABLE;
	else
		if(passable)
			DP.downType = SOLVABLE;
		else
			DP.downType = NORMAL;
	
	return DP;
}

std::string TreeNode::toString()
{
	std::stringstream str;

	str << nr;

	if(!isLeaf())
	{
		str << "(";
		for(unsigned int vtint = 0; vtint < next.size(); vtint++)
		{
			str << next[vtint]->toString();
			if(vtint < next.size()-1)
				str << ",";
		}
		str << ")";
	}

	return str.str();
}

std::string SwitchingTree::strGetType()
{
	std::stringstream str;

	switch(treeType)
	{
		case 1: { str << "EMPTY"; break; }
		case 2: { str << "FREE"; break; }
		case 3: { str << "NORMAL"; break; }
		case 4: { str << "SOLVABLE"; break; }
		case 5: { str << "UNSOLVABLE"; break; }
		case 6: { str << "DIVISIBLE"; break; }
		default:{ str << "UNKNOWN"; break; }
	}
	
	return str.str();
}

std::string SwitchingTree::toStringPaths()
{
	std::stringstream str;
	
	if(!paths.empty())
	{
		for(unsigned int it=0; it<paths.size(); it++)
		{
			str << "Path " << it << ": {";
			vector<unsigned int> itpath = paths[it];
			for(unsigned int jt = 0; (jt+1)<itpath.size(); jt++)
				str << itpath[jt] << ",";
		
			str << itpath[itpath.size()-1] << "}\n";
		}
	}
	return str.str();
}

void SwitchingTree::identifyType()
{
	if(root==0) 
	{
		treeType = EMPTY;
		return;
	}

	if(root->isLeaf()) 
	{
		treeType = FREE;
		return;
	}
	
	if(!identified)
	{
		// analysiere rekursiv den Baum von der root aus.
		downProtocol treeDownProtocol = root->downAnalyze();
		// merke dir den Typ des Baums.
		treeType = treeDownProtocol.downType;
		// speichere alle paths der leaves
		if(treeType == SOLVABLE || treeType == DIVISIBLE)
			for(unsigned int ileaf = 0; ileaf<treeDownProtocol.leaves.size(); ileaf++)
			{
				TreeNode* currentLeaf = treeDownProtocol.leaves[ileaf];
				if(currentLeaf->nr == root->nr)
					paths.push_back(currentLeaf->path);
			}
		identified = true;
	}
}

DLHandler::DLHandler()
{
	_aimedConfig = 0;
	_allunsolvable = false;
	_graphTable.clear();
	_bufferOfIndex.clear();
}

bool DLHandler::Critical()
{
	// return true;
	// Wenn critical-Markierung nicht (mehr) gesetzt, dann kann
	// das Netz nicht wieder kritisch geworden sein.
	
	bool return_val = false;
	
	if(!critical)
		return false;
	
	ComponentVector Buffers = _aimedConfig->Components(Component::BUFFER);
	for(ComponentVector::iterator it1 = Buffers.begin();it1!=Buffers.end();it1++)
	{
		Buffer* it1Buffer = (Buffer*)(*it1);
		
		if(it1Buffer->maeander) return true; 
		// wenn einer schon als MaeanderBuffer erkannt wurde, dann ist sowieso kritisch.
		// neue Buffer mssen (und sollten) nicht mehr maeander-vermerkt werden.
		
		if(!it1Buffer->BQ->empty())
		{
			const Bitmask* frontflitBM = &it1Buffer->BQ->front()->Owner()->getBitmask();		
			const Bitmask* matchBM = _aimedConfig->Coroutine()->routing()
				->validOutput(0,frontflitBM,it1Buffer->In()[0]);
			
			// Wenn ein Puffer ein Paket enthlt, das dort nicht hin wollte
			// und es nicht von einer Source kam
			if(matchBM==0 && it1Buffer->In()[0]->From()->rtti()!=Component::SOURCE && it1Buffer->BQ->front()->head())
			{
				it1Buffer->maeander = true;
				if(Simulation::Instance()->show_actions)
					std::cout << "Signed buffer " << it1Buffer->Index() << " maeander. Proof: " 
						<< it1Buffer->maeander << "\n";
				return_val = true;
			}
			
		}
	}
	if(return_val) return true;
	// Das Netz ist wieder stabil, was vermerkt werden soll.
	critical=false;
	
	return false;
}

void DLHandler::setTables()
{
	// Nimm jeden Buffer
	
	ComponentVector Buffers = _aimedConfig->Components(Component::BUFFER);
	_bufferOfIndex.resize(Buffers.size());
	_graphTable.resize(Buffers.size());
		
	for(ComponentVector::iterator it1 = Buffers.begin();it1!=Buffers.end();it1++)
	{		
		Buffer* it1Buffer = (Buffer*)(*it1);
		Node::Filter* nf = new TriggerNode::Filter();
		NodeVector nv = it1Buffer->nextNodes(nf);
		_graphTable[it1Buffer->Index()].clear();
		
		// setze den Pointer in den entsprechenden Tabelleneintrag
		_bufferOfIndex[it1Buffer->Index()] = it1Buffer;
				
		// Wenn ein Paket in diesem Buffer liegt
		if(!it1Buffer->BQ->empty())
		{
			// Wenn head, dann kommt es darauf an, wo berall er hin gehen knnte
			if(it1Buffer->BQ->front()->head())
			{
				const Bitmask* tmpBM = &it1Buffer->BQ->front()->Owner()->getBitmask();
				Bitmask* resBM;
				// Bedingung, dass Eintrge gemacht werden: Wenn alle Nachfolger volle Buffer sind.
				// nimm jeden Nachfolger des Buffers
				for(NodeVector::iterator itnv=nv.begin();itnv!=nv.end();itnv++)
				{
					// prfe, ob Paket in Buffer dort hin will
					resBM = _aimedConfig->Coroutine()->routing()->validOutput(0,tmpBM,(*itnv)->In()[0]);
					if(resBM!=0)
					{
						// wenn ein Nachfolger nicht ein Buffer ist oder nicht voll, dann kein Eintrag
						if(!((*itnv)->rtti()==Component::BUFFER && ((Buffer*)(*itnv))->BQ->freeSpace()==0))
						{
							_graphTable[it1Buffer->Index()].clear();
							break;
						}
						// Merke dir Eintrag
						_graphTable[it1Buffer->Index()].push_back((*itnv)->Index());
					}
				}				
			}
			// Wenn nicht head, dann gibt es auf jeden Fall einen Eintrag,
			// nmlich dort wo das Flit hin muss. (Nur fr PcT)
			else if(Simulation::Instance()->Active()->Coroutine()->isPCT)
			{
				unsigned int packetNr = it1Buffer->BQ->front()->Owner()->number;
				for(NodeVector::iterator itnv=nv.begin();itnv!=nv.end();itnv++)
				{
					if((*itnv)->rtti()!=Component::BUFFER) continue;
					Buffer* nvBuf = ((Buffer*)(*itnv));
					if(nvBuf->BQ->empty()) continue;
					// wenn ein flit die gleiche Nummer sieht an einem nicht-Tail, dann Eintrag
					if(!nvBuf->BQ->back()->tail() && nvBuf->BQ->back()->Owner()->number==packetNr)
						_graphTable[it1Buffer->Index()].push_back((*itnv)->Index());
				}
			}
		}
	}
}

std::string DLHandler::toStringGraphTable()
{
	std::stringstream str;
	
	bool writtenHead = false;

	for(unsigned int gtit = 0; gtit<_graphTable.size(); gtit++)
	{
		if(!_graphTable[gtit].empty())
		{
			if(!writtenHead)
			{
				writtenHead = true;
				str << "Index | Dependency List\n";
				str << "======|================\n";
			}
			str << gtit << "   ";
			if(gtit<100)	
				str << " ";
			if(gtit<10)
				str << " ";
			str << "| ";
			for(unsigned int lisit = 0; lisit<_graphTable[gtit].size(); lisit++)
			{
				str << _graphTable[gtit][lisit];
				if(lisit<_graphTable[gtit].size()-1)
					str << ", ";
			}
			str << "\n";
		}
	}
	if(!writtenHead)
		str << "Graph-table empty.\n";
	return str.str();
}


void DLHandler::setTrees()
{
	// fr alle Buffer (die umkehren mssen) baue den Switching-Tree auf.
	_trees.clear();
	// Ich nehme zuerst eine lsbare Situation an
	_allunsolvable = false;
	bool exists_dl = false;
	bool exists_solvable_dl = false;
		
	ComponentVector Buffers = _aimedConfig->Components(Component::BUFFER);
	for(ComponentVector::iterator it1 = Buffers.begin();it1!=Buffers.end();it1++){
		Buffer* it1Buffer = (Buffer*)(*it1);
				
		// wir produzieren hier nur switching trees von nichtleeren Buffern.
		if(!it1Buffer->BQ->empty())
		{
			const Bitmask* tmpBM = &it1Buffer->BQ->front()->Owner()->getBitmask();	

			// Wenn das Paket mit Bitmaske tmpBM laut dem Routing nicht in den Buffer htte gehen drfen, 
			// dann ist es auf einem Irrweg. Nur dann bauen wir den switching tree auf.
						
			Bitmask* it1Mask = _aimedConfig->Coroutine()->routing()->validOutput(0,tmpBM,it1Buffer->In()[0]);
			if(it1Mask==0 && it1Buffer->In()[0]->From()->rtti()!=Component::SOURCE)
			{
				TreeNode* TNtmp = new TreeNode(it1Buffer->Index(),0);
				addNodes(TNtmp);
				SwitchingTree* STtmp = new SwitchingTree(TNtmp);
				STtmp->identifyType();
				
				// Merke dir den Baum nur dann, wenn er SOLVABLE, UNSOLVABLE oder DIVISIBLE.
				if(STtmp->getType()>=SOLVABLE) 
				{
					exists_dl = true;
					if(STtmp->getType()!=UNSOLVABLE)
						exists_solvable_dl = true;
					_trees.push_back(STtmp);
				}
				else
					delete STtmp;
			}
		}
	}
	// Falls DL existieren, aber nur unlsbare, dann ist alles unlsbar.
	if(exists_dl && !exists_solvable_dl)
		_allunsolvable = true;
}

std::string DLHandler::toStringTrees(bool with_paths)
{
	std::stringstream str;
	
	if(_trees.size()>0) str << "Set of Trees in Tick " << Simulation::Instance()->Tick() << ":\n";

	for(unsigned int trit=0; trit<_trees.size(); trit++)
	{
		_trees[trit]->identifyType();
		str << _trees[trit]->strGetType() << "-Tree " << trit << ": " << _trees[trit]->getRoot()->toString() << "\n";
		if(with_paths)
			str << _trees[trit]->toStringPaths();
	}
	
	return str.str();
}

bool DLHandler::forceSwitching()
{	
	if(_trees.empty())
		return false;

	unsigned int bufferNum = _aimedConfig->Components(Component::BUFFER).size();

	// Wir fhren eine Liste mit, die alle schon in einem Pfad verwendeten Puffer vermerkt.
	vector<bool> metBuffer;

	// Das ist die iterativ zu fllende Menge der disjunkten Pfade
	vector< vector<unsigned int> > djPaths;

	// Die Liste der Buffer soll zuerst leer sein.
	for(unsigned int i=0; i<bufferNum; i++)
		metBuffer.push_back(false);

	// Fr jeden Baum
	for(unsigned int itree=0; itree<_trees.size(); itree++)
	{
		// hole die Pfade aus dem Baum
		vector< vector<unsigned int> > iPaths = _trees[itree]->getPaths();
		// und fr jeden Pfad
		for(unsigned int jpath=0; jpath<iPaths.size(); jpath++)
		{
			vector<unsigned int> ijPath = iPaths[jpath];
			bool disjoint=true;
			//prfe, ob kein Pfadelement schon in der Bufferliste.
			for(unsigned int k=0; k<ijPath.size(); k++)
			{
				unsigned int ijkBuf = ijPath[k];
				if(metBuffer[ijkBuf]) { disjoint=false; break; }
			}
			// wenn ja, aktualisiere Bufferliste und djPaths-Liste
			if(disjoint)
			{
				for(unsigned int k=0; k<ijPath.size(); k++)
					metBuffer[ijPath[k]] = true;
				djPaths.push_back(ijPath);
			}
		}
	}
	// dies ging O(#trees * #paths * pathlength)	
	// jetzt haben wir disjunkte Pfade in djPaths-Liste.

	// Wende jeden disjunkten Pfad auf die Deadlock-Situation an.
	// Bufferqueues fr Pfad-Buffer auf infinite setzen.
	
	for(unsigned int ipath=0; ipath<djPaths.size(); ipath++)
	{
		vector<unsigned int> ithPath = djPaths[ipath];

		// Setze den Buffer jedes Path-Elements auf infinite

		for(unsigned int kbuf = 0; kbuf<ithPath.size(); kbuf++)
		{
			_bufferOfIndex[ithPath[kbuf]]->BQ->infinite = true;
			if(Simulation::Instance()->show_actions)
				std::cout << "Signed buffer " << ithPath[kbuf] << " (=" 
					<< _bufferOfIndex[ithPath[kbuf]]->Index() << ") infinite. Proof: " 
					<< _bufferOfIndex[ithPath[kbuf]]->BQ->infinite << "\n";
		}
	}
	
	// std::cout << "\n";
	
	return true;
}

void DLHandler::addNodes(TreeNode* where)
{
	// Nimm Tabelleneintrag von where
	
	assert(_graphTable.size()>where->nr);
	
	
	vector<unsigned int> nextlist = _graphTable[where->nr];
	int passedpoint = 0;
	int passedindex = 0;
	int passed = 0;
	bool foundNew = false;
	bool foundOld = false;

	// Gehe durch Eintrge durch
	for(unsigned int i = 0;i<nextlist.size();i++)
	{
		// Wenn der Eintrag nicht schon mal passiert wurde
		passed = where->passedFor(nextlist[i]);
		if(!passed)
		{
			// dann hnge einen Knoten an und fahre iterativ fort. ("Tiefen-Bau")
			TreeNode* TNtmp = new TreeNode(nextlist[i],where);
			where->next.push_back(TNtmp);
			addNodes(TNtmp);
			foundNew = true;
		}
		// Wenn schon passiert, dann merke dir weiteste Schrittweite.
		else if(passedpoint<passed)
		{
			foundOld = true;
			passedpoint = passed;
			passedindex = nextlist[i];
		}
	}
	// Wenn kein Knoten hinzugefgt, so fge (nicht iterativ) den Knoten
	// mit der weitesten Schrittweite hinzu, wenn es einen gibt, der schon
	// passiert wurde.
	if(!foundNew && foundOld) where->next.push_back( new TreeNode(passedindex,where) );
}

