#include <analysis.h>
#include <macros.h>
#include <config.h>

#include <analysis/mathematics.h>

#include <cassert> 
#include <cmath>

#if COMPILE_WITH_MRIP
#include <mpi.h>
#endif


Analysis::VariableInfo* Analysis::_list = NULL;
Analysis::Result* Analysis::_results = NULL;
unsigned int Analysis::_declared = 0;
unsigned int Analysis::_finished = 1;
bool Analysis::_gathering = false;

Analysis* Analysis::_instance = NULL;
string Analysis::_factoryid = "CINSimAnalysis";

#if COMPILE_WITH_MRIP

static int rank;
static int size;
static char* buffer;

static Analysis::Result* globalResults;
static Analysis::Result** localResults;
static int* updates;

//
// Groesse des Sendepuffers
//
#define _DATA_COUNT_ 6
#define _BUFFER_SIZE_ (1000 * (sizeof(double) * _DATA_COUNT_))

#endif




void Analysis::setFactoryId(const string& id)
{
	if(!_instance)
	{
		_factoryid = id;
	}
}


bool Analysis::createInstance()
{
	if(_instance)
		delete _instance;
	
	_instance = (Analysis*)FactoryElement::Factory::self().create(_factoryid);
	
	if(!_instance)
		_ERROR("Failed to create instance for id " << _factoryid << "!");

	return true;
}


void Analysis::Declare(const VariableInfo* list, unsigned int count)
{
	assert(_instance);

	if(!count) 
		_ERROR("Expected at least one variable to be defined!");

	if(_declared)
		_ERROR("This method should only be called once!");
	
	_list = new VariableInfo[count];
	_results = new Result[count];
	_declared = count;
	_finished = 0;
	for(unsigned int i=0; i<_declared; i++)
	{
		_list[i] = list[i];
		_results[i].confidence = list[i].confidence;
		_results[i].precision = list[i].precision;
	}
	
	_instance->declareVariables(list, count);

#if COMPILE_WITH_MRIP

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	
	buffer = (char* ) malloc(_BUFFER_SIZE_);
	MPI_Buffer_attach(buffer, _BUFFER_SIZE_);
	
	if(rank == 0)
	{
		localResults = new Analysis::Result*[count];
		globalResults = new Analysis::Result[count];
		updates = new int[count];
		for(unsigned int i=0; i<count; i++)
		{
			globalResults[i] = _results[i];
			localResults[i] = new Analysis::Result[size];
			updates[i] = 0;
		}
	}
	
#endif	
}


void Analysis::Advance(unsigned int variable, double value)
{
	assert(_instance);
	assert(_gathering);
	
	Result& result = _results[variable-1];

	if(_instance->advanceVariable(variable, value, result))
	{	
		
#if COMPILE_WITH_MRIP		

		if(rank)
		{
			MPI_Bsend(&result, _DATA_COUNT_, MPI_DOUBLE, 0, variable, MPI_COMM_WORLD);		
		}
		else
		{	
			localResults[variable-1][0] = result;
			updates[variable-1]++;
		}
		
#else

		if(result.acceptable())
		{
			assert(_finished < _declared);
					
			_instance->finishVariable(variable);
			_finished++;
		}

#endif
	
	}
}

void Analysis::Reset()
{
	assert(_instance);
	assert(_list);
	assert(_results);
	
	for(unsigned int i=0; i<_declared; i++)
	{
		_results[i].reset();

#if COMPILE_WITH_MRIP
		
		if(rank == 0)
		{
			updates[i] = 0;
			globalResults[i].reset();

			for(unsigned int j=0; j<size; j++)
			{
				localResults[i][j].reset();
			}
		}
		
#endif	
		
	}

	_instance->declareVariables(_list,_declared);
	_finished = 0;
}


void Analysis::StartGathering()
{
	assert(!_gathering);

#if COMPILE_WITH_MRIP
	
	int message;
	MPI_Status status;

	if(rank > 0)
	{
		MPI_Iprobe(0, MPI_ANY_TAG, MPI_COMM_WORLD, &message, &status);

		while(message == 1)
		{
			MPI_Recv(NULL, 0, MPI_DOUBLE, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
			
			if(status.MPI_TAG || status.MPI_TAG > _declared)
			{
				_instance->finishVariable(status.MPI_TAG);
				_finished++;
			}
			else _ERROR("Message with unexpected tag seen!");

			MPI_Iprobe(0, MPI_ANY_TAG, MPI_COMM_WORLD, &message, &status);
		}
	}
	
#endif
	
	_gathering = true;
}

void Analysis::StopGathering()
{
	assert(_gathering);

#if COMPILE_WITH_MRIP
	
	int message = 0;
	MPI_Status status;

	if(rank == 0)
	{
		MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &message, &status);

		while(message)
		{
			if(status.MPI_TAG || status.MPI_TAG > _declared)
			{
				MPI_Recv(&localResults[status.MPI_TAG-1][status.MPI_SOURCE], _DATA_COUNT_, 
						MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
					
				//std::cerr << "++ DATA (" << status.MPI_TAG << ") from " 
				//	<< status.MPI_SOURCE << std::endl;
			
				updates[status.MPI_TAG-1]++;
			}
			else _ERROR("Message with unexpected tag seen!");

			MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &message, &status);
		}

		for(unsigned int v=0; v<Declared(); v++)
		{
			if(updates[v] > 0)
			{				
				Result& result = globalResults[v];
				Result* local = localResults[v];
	
				assert(!result.acceptable());
				
				result.reset();
				
				for(int i=0; i<size; i++)
				{
					//std::cerr << "++ ESTIMATE (" << v+1 << ") from " 
					//	<< i << ": " << local[i].estimate << std::endl;

					result.chkpnts += local[i].chkpnts;
					result.values += local[i].values;
					result.discarded += local[i].discarded;
					result.estimate += local[i].values * local[i].estimate;
					result.variance += local[i].values * local[i].values * local[i].variance;
					result.freedom += local[i].freedom;
				}

				result.estimate /= result.values;
				result.variance /= (result.values * result.values);
				
				if(result.estimate > 0.0)
				{
					double tdist = T((1.0-result.confidence)/2.0, (int)result.freedom);
					result.error = (tdist * sqrt(result.variance))/result.estimate;
				}
				else 
					result.error = DBL_MAX;

				updates[v] = 0;
		
				//std::cerr << "++ UPDATE (" << v+1 << "): " 
				//	<< ", cnt=" << result.values
				//	<< ", est=" << result.estimate 
				//	<< ", var=" << result.variance 
				//	<< ", err=" << result.error 
				//	<< std::endl;	

				if(result.acceptable())
				{					
					assert(_finished < _declared);
					
					_instance->finishVariable(v+1);
					_finished++;
					
					// keine weiteren updates fuer v
					updates[v] = INT_MIN;
					
					for(int t=1; t<size; t++)
					{
						if(Finished())
						{
							//
							// bei letzter Messung blockierend, um Prozesse aufzusammeln
							//
							MPI_Send(NULL, 0, MPI_DOUBLE, t, v+1, MPI_COMM_WORLD);

							Analysis::Result dummy;
							MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, 
									&message, &status);
							//
							// Empangspuffer leeren
							// 
							while(message)
							{
								MPI_Recv(&dummy, _DATA_COUNT_,MPI_DOUBLE, MPI_ANY_SOURCE, 
											MPI_ANY_TAG, MPI_COMM_WORLD, &status);
					
								MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, 
										&message, &status);
							}
						}
						else
						{
							//
							// ansonsten nicht-blockierend (asynchron) wegen Geschwindigkeit
							//
							MPI_Bsend(NULL, 0, MPI_DOUBLE, t, v+1, MPI_COMM_WORLD);
						}
					}
				}
			}
		}	
	}

	if(Finished()) MPI_Barrier(MPI_COMM_WORLD);

#endif
	
	_gathering = false;
}

const Analysis::Result& Analysis::CurrentResult(unsigned int index)
{
	assert(0 < index && index <= Declared());

#if COMPILE_WITH_MRIP
	
	if(rank == 0)
		return globalResults[index-1];

#endif
	
	return _results[index-1];
}
