#include <initialtransient.h>
#include <spectralmean.h>
#include <mathematics.h>

#include <iostream>
#include <cmath>
#include <cassert>

#define _transient_max_ 10000
#define _heuristic_max_ 1024
#define _crossings_min_ 25

#define _nv_ 100
#define _alpha_ 0.01
#define _kappa_ 7			// C2(nv,d)
#define _confidence_ 0.95

#define MAX(a,b) (unsigned int)((a)>=(b)?(a):(b))

											  
InitialTransient::InitialTransient()
{
	_k = 0;
	
	_state = HEURISTIC;
	advanceInternal=&InitialTransient::applyHeuristic;
	
	_boffset = 0;
	_bsum = 0.0;
	_buffer = new double[_heuristic_max_];

	_tdist = T(_alpha_/2.0, _kappa_);
}

InitialTransient::~InitialTransient()
{
	delete [] _buffer;
}


int InitialTransient::applyHeuristic(double value)
{
	assert(_state == HEURISTIC);

	unsigned int crossings = 0;
	double mean = 0.0;
	
	_buffer[_boffset++] = value;
	_bsum += value;
	mean = _bsum/(double)_boffset;
	
	//std::cerr << _boffset << " " << mean << " " << value << std::endl;
	
	for(unsigned int i=1; i<=_boffset; i++)
	{
		if((_buffer[i-1] < mean && mean < _buffer[i+1])
			|| (mean < _buffer[i-1] && _buffer[i] < mean)
			|| (mean == _buffer[i-1] && mean == _buffer[i]))

			crossings++;
	}

	if(crossings >= _crossings_min_ || _boffset == _heuristic_max_) 
	{
		_k = _boffset;
		_delta_k = (unsigned int)(0.5 * _k);
		
		_nt = MAX(2*_nv_, _delta_k);
		_delta_nt = _nt;

		delete[] _buffer;
		_buffer = new double[_nt];
		_boffset = 0;
		
		_state = TEST;
		advanceInternal = &InitialTransient::applyTest;

		//std::cerr << "Initial Transient (heuristic): " << _k << std::endl;
	}	

	return _state;
}


int InitialTransient::applyTest(double value)
{		
	if(_state != TEST) return _state;

	_buffer[_boffset++] = value;
	
	if(--_delta_nt) return _state;
	
	double variance = SpectralMean::Variance(_buffer+(_boffset-_nv_), _nv_, _confidence_);
	double mean_l = 0.0;
	double mean_sum = 0.0;
	double s1 = 0.0;
	double s2 = 0.0;
	double beta_l = 0.0;
	
	for(unsigned int l=1; l<=_nt; l++)
	{
		beta_l = l - (l*l)/(double)_nt;
		mean_sum += _buffer[l-1];
		mean_l = mean_sum/(double)l;
		
		s1 += beta_l;
		s2 += beta_l * mean_l;
	}

	double T = fabs(sqrt(45)/ (_nt * sqrt( _nt*_nv_*variance)) * (mean_l*s1 - s2));

	//
	// Test der Hypothese
	// 
	if( T <= _tdist )
	{	
		_state = FINISHED;
		advanceInternal = &InitialTransient::applyNop;

		//std::cerr << "Initial Transient (final): " << _k << std::endl;
		
		return _state;
	}

	//
	// Weitere Werte verwerfen: k = k + delta_k
	// 
	_k += _delta_k;

	//
	// Abbruch, wenn Phase zu lang 
	//
	if(_k + _nt >= _transient_max_)
	{
		_state = FAILED;
		advanceInternal = &InitialTransient::applyNop;

		return _state;
	}
	
	//
	// ...oder weiter
	//
	_delta_nt = _delta_k;
	_boffset = _nt-_delta_nt;

	for(unsigned int i=0; i<_boffset; i++)
		_buffer[i] = _buffer[i+_delta_k]; 
	
	return _state;
}


int InitialTransient::applyNop(double)
{
	assert(_state == FINISHED || _state == FAILED);

	return _state;
}


bool InitialTransient::operator>>(Estimation* est) const
{
	assert(_state == FINISHED && !_delta_nt);
	
	bool chkpnt = false;
		
	est->setTransient(_k);

	for(unsigned int i=0; i<_nt; i++)
	{
		chkpnt |= est->advance(_buffer[i]);
	}

	return chkpnt;
}

