#include <spectralmean.h>
#include <mathematics.h>

#include <cassert>
#include <cmath>
#include <cfloat>
#include <iostream>

#define _CHKPNT_MIN_ 512

#define _N_ 100
#define _K_ 25
#define _C1_ 0.882	// C1(K=25,d=2)
#define _C2_ 7 // C2(K=25,d=2)
#define _d_ 2

#define MAX(a,b) ((a) < (b) ? b : a)

SpectralMean::SpectralMean(double confidence) : Estimation(confidence)
{
	_buffer = NULL;
	_buffersize = 2*_N_;
	_offset = 0;
	
	if(_buffersize)
	{
		_buffer = new double[_buffersize];
		_buffer[_offset] = 0.0;
	}

	_batchsize = 1;
	_batchcount = _batchsize;

	_freedom = _C2_;
	_tdist = T((1.0 - _confidence)/2.0, (int)_freedom);

	_fj = new double[_K_];
	for(unsigned int j=0; j<_K_; j++)
	{
		_fj[j] = (4.0*(j+1.0) - 1.0) / (2.0*_N_);
	}
}

SpectralMean::~SpectralMean()
{
	if(_buffer)
		delete [] _buffer;

	delete [] _fj;
}

inline bool SpectralMean::checkPoint() const
{
	int count = ((_batchsize * (_offset+1)) - _batchcount);
	
	return count && (count % MAX(_CHKPNT_MIN_, _batchsize) == 0);
}

void SpectralMean::reset()
{
	assert(_buffer);
	
	Estimation::reset();
	
	_offset = 0;
	_batchsize = 1;
	_batchcount = _batchsize;

	_buffer[_offset] = 0.0;
}

bool SpectralMean::advance(double value)
{	
	double &batchmean = _buffer[_offset];
	
	batchmean += value;
	_batchcount--;
	
	if(!_batchcount)
	{
		batchmean /= (double)_batchsize;
		_estimate = (_offset*_estimate + batchmean)/(_offset+1);
		
		// re-batching
		if(_offset == _buffersize-1)
		{
			unsigned int i=0;
			_estimate = 0.0;
			
			for(; i<_buffersize/2; i++)
			{
				_buffer[i] = 0.5*(_buffer[2*i] + _buffer[2*i+1]);
				_estimate += _buffer[i];
			}
			_estimate /= i;
			
			_offset = _buffersize/2 - 1;
			_batchsize *= 2;
		}
		
		assert(_offset+1 < _buffersize);
		
		_buffer[++_offset] = 0.0;
		_batchcount = _batchsize;
	}

	if(checkPoint())
	{
		assert(_offset>=_N_);
		
		_variance = estimateVariance(_buffer+(_offset-_N_), _N_);

		_error = (_tdist * sqrt(_variance)) / (_estimate > 0.0 ? _estimate : DBL_MIN);

		return true;
	}

	return false;
}

double SpectralMean::estimateVariance(const double *X, unsigned int size) const
{
	unsigned int n=0, j=0;
	
	double P[2*_K_];
	double Lfj[_K_];
	double a[_d_+1];

	assert(_N_ == size);
	
	// P[n-1] = I(n/N), n = 1..2K
	periodograms(2*_K_, _N_, X, P);

	// L(fj)
	for (j=0; j<_K_; j++) 
	{
		double x = (P[2*j] + P[2*j+1])/2.0;	
		
		// Keine Null erlaubt
		if (x == 0)
			x = DBL_MIN;
		
		Lfj[j] = log(x) + 0.270;
	}
	
	// a[0],..,a[d]
	leastSquares(_fj,Lfj,_K_,_d_,a);

	// C1*px0/N
	return (_C1_*exp(a[0]))/_N_;
}

double SpectralMean::Variance(double *data, unsigned int n, double confidence)
{
	SpectralMean sa(confidence);

	return sa.estimateVariance(data, n);
}
