#include <spectralquantile.h>
#include <psquarequantile.h>

#include <cmath>
#include <cassert>
#include <iostream>

#define _Nb_ 512
#define _K_ 30
#define _d_ 2

#define _eps_ 0.001


SpectralQuantile::SpectralQuantile(double confidence, double p) : Estimation(confidence)
{
	_alpha = 1.0 - _confidence;

	_buffer = new double[_Nb_];
	_bsize = _Nb_;
	_bcount = 0;

	_s=1;
	_fj = new double[_K_];
	_h_fj = new double[_K_];
	_csum = new double[4];
	
	for(unsigned int j=0; j<_K_; j++)
	{
		assert(_K_ < (double)_Nb_ / (4.0));
	
		_fj[j] = (4.0*(j+1.0) - 1.0) / (2.0*_Nb_);
		
		// Null vermeiden
		_h_fj[j] = DBL_MIN;
	}
	
	_psq = new PSquareQuantile(p);
}

SpectralQuantile::~SpectralQuantile()
{
	delete [] _buffer;
	delete [] _fj;
	delete [] _h_fj;
	delete [] _csum;
		
	delete _psq;
}

void SpectralQuantile::reset()
{
	Estimation::reset();

	_s = 1;
	_bcount = 0;
	
	for(unsigned int j=0; j<_K_; j++)
	{	
		_h_fj[j] = DBL_MIN;
	}
	
	_psq->reset();
}

bool SpectralQuantile::checkPoint() const
{
	return !_bcount;
}


bool SpectralQuantile::advance(double value)
{
	_psq->advance(value);
	_buffer[_bcount++] = value;
	
	// Checkpoint
	if(_bcount == _bsize)
	{
		unsigned int i, j;
		
		double P[2*_K_];
		double Lfj[_K_];
		double a[_d_+1];
		double m[4];

		// Est(x_p) = x_p,s
		_estimate = _psq->quantile();
		
		// I(j,s;xp), j=1..Nb
		for(j=0; j<_bsize; j++)
		{
			_buffer[j] = (_buffer[j] <= (_estimate + _eps_)) ? 1.0 : 0.0;
		}
		
		// degrees of freedom
		double df = 4.0 * (double)_s;

		// m_i, i=1..4
		m[0] = (1.0 / df) + (1.0 / (3.0*df*df));
		m[1] = 2.0 / (df-1);
		m[2] = -4.0 / ((df-1.0) * (df-1.0));
		m[3] = (8.0 / ((df-1.0)*(df-1.0)*(df-1.0))) + (12.0 / ((df-1.0)*(df-1.0)));

		// P[n-1] = I(n/N), n = 1,2...
		periodograms(2*_K_, _Nb_, _buffer, P);
		
		// h_s(fj), Lfj = log(SUM[h_s(fj)]/s) + m_1
		for(j=0; j<_K_; j++)
		{
			double hs_fj = (P[2*j] + P[2*j+1])/2.0;
			
			_h_fj[j] += hs_fj;
			
			Lfj[j] = log(_h_fj[j] / (double)_s) + m[0];
		}
				
		//
		// Erstes Segment
		//
		if(_s == 1)
		{		
			double C[(_d_+1)*(_K_)];

			// a[0],..,a[d]: y = Xa + e ==> a = Cy
			leastSquares(_fj, Lfj, _K_, _d_, a, C);

			// SUM((C_1,j)^i), i=1..4, j=1..n
			for(i=0; i<4; i++)
			{
				_csum[i] = 0.0;
				for(j=0; j<_K_; j++)
				{
					_csum[i] += pow(C[j], i+1);
				}
			}
		}	
		//
		// alle weiteren Segmente
		//
		else
		{
			// a[0],..,a[d]
			leastSquares(_fj, Lfj, _K_, _d_, a);
		}

		// m_i', i=1..4
		for(i=0; i<4; i++) m[i] *= _csum[i];
		
		// h(0) = e^a_0 / (1 + m_2'/2 + m_3'/6 + m_4'/24)
		double h0 = exp(a[0]) / (1.0 + (m[1]/2.0) + (m[2]/6.0) + (m[3]/24.0));
		
		// Var(x_p) = h(0) / (N * f(x_p)^2)
		double f_xp = _psq->density();
		_variance = h0 / ((double)_s * (double)_Nb_ * f_xp * f_xp);
				
		// ndf = 3 / (3*m_2' + 2*m_3' + m_4')
		_freedom = rint(3.0 / (3.0*m[1] + 2.0*m[2] + m[3]));
		
		// Err(x_p) = (t(1-alpha/2,ndf) * Var(x_p)^0.5) / Est(x_p)
		_error = (T(_alpha/2.0, (int)_freedom) * sqrt(_variance)) / 
									(_estimate > 0.0 ? _estimate : DBL_MIN);
		

		//
		// naechstes Segment beginnen
		//
		_bcount = 0;
		_s++;

		return true;
	}

	return false;
}

