#ifndef __TEST_ESTIMATION_H__
#define __TEST_ESTIMATION_H__

#include <cppunit/extensions/HelperMacros.h>

using namespace CppUnit;

class Estimation;
class TestDistribution;

//! Unit-Tests fuer Abschaetzungen (Basisklasse).
/*!
 * Tests anhand von Verteilungen mit bekannten Mittelwerten (mu) und
 * p-Quantilen (xp).
 * 
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class Test_Estimation : public TestFixture
{
public:

	//! Implementierung der Klasse TestFixture.
	void setUp();

	//! Implementierung der Klasse TestFixture.
	void tearDown();

	//! Test mit Gleichverteilung [0,1].
	/*!
	 * mu = 0.5
	 * xp = p
	 */
	void uniform_check();

	//! Test mit Exponentialverteilung.
	/*!
	 * mu = 1/lambda
	 * xp = -1/lambda*ln(1-p)
	 */
	void exponential_check();

protected:
	
	//! Erzeugung einer Instanz einer Abschaetzung.
	virtual Estimation* create() { return NULL; }
	
	//! Anwendung einer Testverteilung auf Abschaetzungsmethode.
	/*!
	 * Es wird geprueft, ob Methode den tatsaechlichen Wert mit
	 * vorgegebener Genauigkeit und Konfidenz abschaetzt. Dazu wird
	 * eine hinreichend grosse Anzahl von Durchlaeufen gestartet.
	 */
	void applyDistribution(const TestDistribution *td);
	
	//! Untersuchte Instanz einer Abschaetzungsmethode.
	Estimation *_est;

private:

	//! Aktueller Fehler.
	double _error;

	//! Maximale Anzahl von fehlerhaften Abschaetzungern.
	unsigned int _fmax;
};


//! Test der Mittelwertsbaschaetzung mit Standard-Methode.
class Test_IndependentMean : public Test_Estimation
{
CPPUNIT_TEST_SUITE( Test_IndependentMean );
CPPUNIT_TEST( uniform_check );
CPPUNIT_TEST( exponential_check );
CPPUNIT_TEST_SUITE_END();

protected:

	/*!
	 * Implementierung der Klasse Test_Estimation.
	 *
	 * @return Instanz der Klasse IndependentMean.
	 */
	Estimation* create();
};


//! Test der Mittelwertabschaetzung mit Spektral-Methode.
class Test_SpectralMean : public Test_Estimation
{
CPPUNIT_TEST_SUITE( Test_SpectralMean );
CPPUNIT_TEST( uniform_check );
CPPUNIT_TEST( exponential_check );
CPPUNIT_TEST_SUITE_END();

protected:

	/*! 
	 * Implementierung der Klasse Test_Estimation.
	 *
	 * @return Instanz der Klasse SpectralMean.
	 */
	Estimation* create();
};


//! Test der Quantilen-Abschaetzung mit Spektral-Methode.
class Test_SpectralQuantile : public Test_Estimation
{
CPPUNIT_TEST_SUITE( Test_SpectralQuantile );
CPPUNIT_TEST( uniform_check );
CPPUNIT_TEST( exponential_check );
CPPUNIT_TEST_SUITE_END();

protected:

	/*! 
	 * Implementierung der Klasse Test_Estimation.
	 *
	 * @return Instanz der Klasse SpectralQuantile.
	 */
	Estimation* create();
};

#endif
