#include <cinsimenv.h>
#include <sequencer.h>
#include <macros.h>

#include <parser/cmdparser.h>

#include <analysis.h>
#include <prng.h>

#include <cassert>
#include <fstream>
#include <csignal>

#if COMPILE_WITH_MRIP
#include <mpi.h>
#endif


#define _TAG_ 0

REGISTER_ELEMENT(CINSimEnv, "CINSimEnv")

using namespace std;

#if COMPILE_WITH_MRIP
int rank;
int size;
streambuf* oldcout;
streambuf* oldcerr;
ofstream devnull; 
#endif


bool CINSimEnv::initialize(int argc, char** argv)
{
#if COMPILE_WITH_MRIP
	MPI_Status status;
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	if(rank)
	{
		oldcout = cout.rdbuf();
		oldcerr = cerr.rdbuf();

		devnull.open("/dev/null");

		assert(devnull.is_open());
		
		cout.rdbuf(devnull.rdbuf());
		cerr.rdbuf(devnull.rdbuf());
	}
		
	if (signal(SIGUSR2, Sequencer::SignalHandler) == SIG_ERR)
		_WARNING("Unable to set signal handler SIGUSR2!");

	std::cout << "MPI-Mode: Use 'kill -USR2 " << (int) getpid () << "' for intermediate results, "
		<< "ctrl-c for aborting" << std::endl;

#else
	
	if (signal(SIGINT, Sequencer::SignalHandler) == SIG_ERR)
		_WARNING("Unable to set signal handler for ctrl-c events!");

#endif
	
	//
	// Kommandozeile, Konfigurationsdatei
	//
	if(!Environment::initialize(argc, argv))
		return false;
	
	//
	// Stochastische Auswertung
	// 
	Analysis::setFactoryId(getValue(_ANALYSIS_ID_));
	Analysis::createInstance();
	assert(Analysis::Instance());
	
	//
	// Generator fuer Zufallszahlen
	// 
	unsigned int stream = 0;
	if(isSet(_PRNG_STREAM_))
	{	
		stream = (uint)getValue(_PRNG_STREAM_);
	}
	else
	{
#if COMPILE_WITH_MRIP
		if(rank == 0) 
		{
#endif		
			//FIXME: Alternative fuer Plattformen ohne /dev/random ?!

			FILE *devrandom;
			if ((devrandom = fopen("/dev/random","r")) == NULL) 
			{
				_WARNING("Cannot open /dev/random, setting stream to 0\n");
			} 
			else 
			{
				fread(&stream, sizeof(stream), 1, devrandom);
				fclose(devrandom);
			}
		}
#if COMPILE_WITH_MRIP
	}	
	if(rank == 0)
	{
		// Share ID
		for(unsigned int i=1; i<size; i++)
			MPI_Send(&stream, 1, MPI_UNSIGNED, i, _TAG_, MPI_COMM_WORLD);
	}	
	else
	{			
		// Receive ID
		MPI_Recv(&stream, 1, MPI_UNSIGNED, 0, _TAG_, MPI_COMM_WORLD, &status);
	
		stream += rank;
	}
#endif
	
	// Eventuell neuer Stream
	char tmp[sizeof(unsigned int)*8+1];
	sprintf(tmp, "%u", stream);
	addValue(_PRNG_STREAM_, ValueString(tmp));
	
	PRNG::setStreamId(stream);
	PRNG::setFactoryId(getValue(_PRNG_ID_));
	PRNG::createInstance();	
	
	assert(PRNG::Instance());

	return true;
}

bool CINSimEnv::finalize()
{	
#if COMPILE_WITH_MRIP	
	int initialized;
	MPI_Initialized(&initialized);
	if(initialized)
	{
		MPI_Finalize();
		if(rank)
		{
			cout.rdbuf(oldcout);
			cerr.rdbuf(oldcerr);

			devnull.close();
		}
	}
#endif
	
	return true;
}


CmdParser* CINSimEnv::createCmdParser() const
{
	return new CINSimCmdParser;
}
