#include <cinparser.h>
#include <errhandler.h>
#include <environment.h>
#include <macros.h>

#include <utility/valuestring.h>

#include <string>
#include <iostream>
#include <cassert>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/internal/XMLScanner.hpp>
#include <xercesc/framework/XMLPScanToken.hpp>
#include <xercesc/framework/psvi/XSElementDeclaration.hpp>

using std::cout;
using std::endl;
using std::string;

CINParser::CINParser()
{    
	try 
	{	
		// Initialisierung von Xerces-C
		XMLPlatformUtils::Initialize();
	}
		
	catch (const XMLException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.getMessage());
		cout << "Failed to set up Xerces-C++ for parsing network file! " << endl;
		cout << "Message: " << message << endl;
		XMLString::release(&message);
		exit(1);
	}
	
	string schema = Env()->getValue("XML_SCHEMA");
	string ns = Env()->getValue("XML_NAMESPACE");

	if (schema == snull || ns == snull) 
	{
		_ERROR("Missing schema and schema namespace settings! " 
				<< "Use -n and -s switch or a configuration file.");
	} 
	else 
	{
		string schemadef = ns + " " + schema;
		setExternalSchemaLocation( schemadef.c_str());  
	}

	// notwendig fuer Zugriff auf Typen
	setDoSchema(true);
	setDoNamespaces(true);
  
	// Abbruch bei erstem Fehler
	setExitOnFirstFatalError((int)Env()->getValue(_XML_STOPONERROR_));
	setValidationConstraintFatal(getExitOnFirstFatalError());
	
	if (((string)Env()->getValue(_XML_PROCESSING_)) == _PROCESSING_EXT_)
	{
		// Default fuer Schema-Validierung     	  
		setDoValidation(true);
		setValidationScheme(Val_Always);
		setValidationSchemaFullChecking(true);
	
		// Parse-Methode setzen (Progressive)
		parseMethod = &CINParser::parseProgressive;  
	}

	else if (((string)Env()->getValue(_XML_PROCESSING_)) == _PROCESSING_SCHEMA_) 
	{
		// Default fuer Schema-Validierung
		setDoValidation(true);
		setValidationScheme(Val_Always);
		setValidationSchemaFullChecking(true);

		// Parse-Methode setzen (Standard)
		parseMethod = &CINParser::parse;
	}

	else if ((string)Env()->getValue(_XML_PROCESSING_) == _PROCESSING_MIN_) 
	{    
		setDoValidation(false);
		setValidationScheme(Val_Auto);      

		// Parse-Methode setzen (Standard)
		parseMethod = &CINParser::parse;
	}

	else
		_ERROR("Failed to determine the XML validation procedure!");
		
	// Fehlerbehandlung
	setErrorHandler(new ErrHandler());

	// Hmmm, keine Ahnung warum, aber ohne SIGSEGV
	cacheGrammarFromParse(true);
  
	// PSVI-Handler zum sammeln von Schema-Informationen.
	setPSVIHandler(new PSVIHandlerImpl(_elements));
}


bool CINParser::parseProgressive(const char* xmlFile)
{
	XMLPScanToken token;
	bool parsing;

	try 
	{
		// progressives parsen  
		parsing = parseFirst(xmlFile, token);

		while(parsing)
		{
			//
			// hier koennen Constraints fuer XML-Files geprueft werden
			//

			parsing = this->parseNext(token);
		}    
	}  

	catch (const XMLException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.getMessage());
		cout << "Exception message is:" << endl << message << endl;
		XMLString::release(&message);
		return false;
	}
	catch (const DOMException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.msg);
		cout << "Exception message is:" << endl << message << endl;
		XMLString::release(&message);
		return false;
	} 
	catch (...)
	{
		cout << "Unexpected Exception \n" ;
		return false;
	}
	return !((ErrHandler*)getErrorHandler())->sawErrors();
}
  
bool CINParser::parse(const char* xmlFile)
{
	try
	{
		XercesDOMParser::parse(xmlFile);
	}
	catch (const XMLException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.getMessage());
		cout << "Exception message is:" << endl << message << endl;
		XMLString::release(&message);
		return false;
	}
	catch (const DOMException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.msg);
		cout << "Exception message is:" << endl << message << endl;
		XMLString::release(&message);
		return false;
	} 
	catch (...)
	{
		cout << "Unexpected Exception" << endl;
		return false;
	}
	return !((ErrHandler*)getErrorHandler())->sawErrors();
}

void CINParser::error( const unsigned int                code
                     , const XMLCh* const                msgDomain
                     , const CINParser::ErrTypes         errType
                     , const XMLCh* const                errorText
                     , const XMLCh* const                systemId
                     , const XMLCh* const                publicId
                     , const XMLSSize_t                  lineNum
                     , const XMLSSize_t                  colNum)
{	
	SAXParseException toThrow = SAXParseException
	(
          errorText
        , publicId
        , systemId
        , lineNum
        , colNum
        , getMemoryManager()
	);

	if(!getErrorHandler())
	{
		if (errType == CINParser::ErrType_Fatal)
		throw toThrow;
		return;
	}

	switch(errType)
	{
		case CINParser::ErrType_Warning:
			getErrorHandler()->warning(toThrow);  
			break;

		case CINParser::ErrType_Error:
			getErrorHandler()->error(toThrow);
			break;

		case CINParser::ErrType_Fatal:
			getErrorHandler()->fatalError(toThrow);
			break;

		case CINParser::ErrType_Check:
			((ErrHandler*)getErrorHandler())->checkError(toThrow);
			break;

		default:
			((ErrHandler*)getErrorHandler())->unknownError(toThrow); 
	}
}

int CINParser::getParseErrors() const
{
	return ((ErrHandler*)getErrorHandler())->getErrCount();     
}

void CINParser::resetErrors()
{
	((ErrHandler*)getErrorHandler())->resetErrors();
}

void CINParser::PSVIHandlerImpl::handleElementPSVI(const XMLCh* const name, 
		const XMLCh* const uri, 
		PSVIElement *elementInfo)
{
	assert(name);
	assert(elementInfo);
	
	if(!elementInfo->getElementDeclaration()) return;
		
	assert(XMLString::equals(name, elementInfo->getElementDeclaration()->getName()));

	_map[XMLString::transcode(name)] = elementInfo->getElementDeclaration();
}
