#include <cmdparser.h>
#include <macros.h>
#include <utility/valuestring.h>

#include <iostream>
#include <iomanip>
#include <csignal>
#include <list>

using namespace std;

bool CmdParser::parseArgs(int argc, char** argv)
{
	int myopt;

#ifdef HAVE_GETOPT_LONG
	int option_index = 0;
	while ((myopt = getopt_long(argc, argv, Flags(), LongOpts(), &option_index)) != -1)
	{
#else
    while ((myopt = getopt(argc, argv, Flags())) != -1) 
	{
#endif
		/* parse each option */
		switch (myopt) 
		{			
			// XML: Validierung (Schema)
			case 'c':
				addValue(_XML_PROCESSING_, ValueString(_PROCESSING_SCHEMA_));
				break;
			// Data file
			case 'd':
				addValue(_DATA_FILE_, ValueString(optarg));
				break;
			// XML: Erweiterte Validierung (Schema + Constraints)
			case 'e':
				addValue(_XML_PROCESSING_, ValueString(_PROCESSING_EXT_));
				break;
			// optionale Konfigurationsdatei
			case 'f':
				addValue(_CONF_FILE_, ValueString(optarg));
				break;
			// usage
			case 'h':
				addValue(_SHOW_USAGE_, ValueString("1"));
				break;
			// Nice level
			case 'i':
				if(!ValueString(optarg).isNumber())
					_ERROR("Invalid nice level: " << optarg)
				addValue(_NICE_LEVEL_, ValueString(optarg));
				break;
			// plugins
            case 'l':
				addValue(_SHOW_PLUGINS_, ValueString("1"));
				break;
			// XML: Minimaler Check (Syntax) 
			case 'm':
				addValue(_XML_PROCESSING_, ValueString(_PROCESSING_MIN_));
				break;
			// Namespace
			case 'n':
				addValue(_XML_NAMESPACE_, ValueString(optarg));
				break;
			// Gnuplot
			case 'p':
				addValue(_PLOT_FILE_, ValueString(optarg));
				break;
			// Quiet, kein Fortschritt anzeigen	
			case 'q':
				addValue(_SHOW_PROGRESS_, ValueString("0"));
				break;
			// optionales XML-Schema
			case 's':
				addValue(_XML_SCHEMA_, ValueString(optarg));
				break;
			// Timer
			case 't':
				addValue(_SHOW_TIME_, ValueString("1"));
				break;
			// Verbose	
			case 'v':
				addValue(_SHOW_VERBOSE_, ValueString("1"));
				break;

			// Confidence level
			case 'C':
				if(!ValueString(optarg).isNumber())
					_ERROR("Invalid confidence level: " << optarg);
				addValue(_CONFIDENCE_, ValueString(optarg));
				break;
			// Generator fuer Zufallszahlen
			case 'G':
				if(strcmp(optarg,"MTwister") && strcmp(optarg,"CMRG") && strcmp(optarg,"Rand48"))
					_ERROR("Invalid random number generator: " << optarg)
				addValue(_PRNG_ID_, ValueString(optarg));
				break;
			// Engines
			case 'E':
				if(!ValueString(optarg).isNumber())
					_ERROR("Invalid number of engines: " << optarg);
				addValue(_ENGINES_, ValueString(optarg));
				break;
			// Precision
			case 'P':
				if(!ValueString(optarg).isNumber())
					_ERROR("Invalid precision: " << optarg);
				addValue(_PRECISION_, ValueString(optarg));
				break;
			// Repetitions
			case 'R':
				if(!ValueString(optarg).isNumber())
					_ERROR("Invalid number of repetitions: " << optarg);
				addValue(_REPETITIONS_, ValueString(optarg));
				break;
			// random number stream
			case 'S':
				if(!ValueString(optarg).isPositiveNumber())
					_ERROR("Invalid random number stream: " << optarg)
				addValue(_PRNG_STREAM_, ValueString(optarg));
				break;			// Variable(n)
			case 'V':
				parseVariables(optarg);
				break;
				
			// Abbruch bei erster unbekannter option
			case '?':
				exit(1);
				break;
				
			// Fallback fuer fehlenden case
			default:
				_ERROR("Missing case for valid option! " << myopt);
				break;
		} // switch
	} // while

	if ( argc > optind+1)
		_ERROR("To many non-option arguments!");

	/* check for a single additional argument */
	if (optind < argc) 
	{
		while (optind < argc)
			addValue(_XML_INPUT_, ValueString(argv[optind++]));
	}
	return true;
}

void CmdParser::printUsage(int argc, char** argv) const 
{	
	const char* flags = Flags();
	
	string pname = argv[0];
	string seperators = "/-";
	list<string> fragments;
	split(pname,seperators,fragments);
	
	cout << left;
	
	cout << "CINSim v" << VERSION << " -- " << Info() << endl;
	
	cout << "Usage: " << fragments.back() << " [option]... [xml-file]" << endl;
	cout << endl << "Valid options are:" << endl;

	if(strchr(flags,'v'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -v," << setw(25) << "--verbose";
		cout << "Print verbose messages" << endl;
		cout << setw(42) << "" << "(only visible when not running in background)";
		cout << endl;
	#else
		cout << setw(20) << "    -v" << "Print verbose messages" << endl;
		cout << setw(20) << "" << " (only visible when not running in background)";
		cout << endl;
	#endif
	}
	if(strchr(flags,'q'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -q," << setw(25) << "--quiet";
	#else
		cout << setw(20) << "    -q";
	#endif
		cout << "Do not print state messages" << endl;
	}
	if(strchr(flags,'h'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -h," << setw(25) << "--help";
	#else
		cout << setw(20) << "    -h";
	#endif
		cout << "Print this message and terminate" << endl;
	}
	if(strchr(flags,'t'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -t," << setw(25) << "--timer";
	#else
		cout << setw(20) << "    -t";
	#endif	
		cout << "Show execution time" << endl;
	}
	if(strchr(flags,'f'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -f FILE," << setw(25) << "--config-file=FILE";
	#else
		cout << setw(20) << "    -f FILE";
	#endif
		cout << "Specify the configuration file" << endl;
	}
	if(strchr(flags,'s'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -s FILE," << setw(25) << "--schema=FILE";
	#else
		cout << setw(20) << "    -s FILE";
	#endif
		cout << "Specify the XML schema file" << endl;
	}
	if(strchr(flags,'n'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -n URL," << setw(25) << "--namespace=URL";
	#else
		cout << setw(20) << "    -n URL";
	#endif
		cout << "Specify the XML namespace url" << endl;
	}
	if(strchr(flags,'m'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -m," << setw(25) << "--minimal-validation";
	#else
		cout << setw(20) << "    -m";
	#endif
		cout << "Only minimal validation (syntax checking)" << endl;
	}
	if(strchr(flags,'c'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -c," << setw(25) << "--schema-validation";
	#else
		cout << setw(20) << "    -c";
	#endif
		cout << "Validation against schema (default) " << endl;
	}
	if(strchr(flags,'e'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -e," << setw(25) << "--extended-validation";
	#else
		cout << setw(20) << "    -e";
	#endif
		cout << "Validation against schema and constraints" << endl;
	}
	if(strchr(flags,'i'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -i LEVEL," << setw(25) << "--nice-level=LEVEL";
	#else
		cout << setw(20) << "    -i LEVEL";
	#endif
		cout << "Specify the nice level; 0-19 (user), <0 (superuser)" << endl;
	}
	if(strchr(flags,'d'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -d FILE," << setw(25) << "--data-file=FILE";
	#else
		cout << setw(20) << "    -d FILE";
	#endif
		cout << "Write results to FILE" << endl;
	}	
	if(strchr(flags,'p'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -p FILE," << setw(25) << "--plot-file=FILE";
	#else
		cout << setw(20) << "    -p FILE";
	#endif		
		cout << "Create a plot (postscript) with gnuplot" << endl;
	}
	if(strchr(flags,'l'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -l," << setw(25) << "--plugin-list";
	#else		
		cout << setw(20) << "    -l";
	#endif
		cout << "Show all plugins and terminate" << endl;
	}
	if(strchr(flags,'V'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -V ID=VALUE" << setw(25) << "--variable ID=VALUE"; 
		cout << "Assign VALUE to variable ID" << endl;
		cout << setw(42) << "" << "(also semicolon-separated list in double quotes)";
		cout << endl;
	#else		
		cout << setw(20) << "    -V ID=VALUE" << "Assign VALUE to variable ID" << endl;
		cout << setw(21) << "" << "(also semicolon-separated list in double quotes)";
		cout << endl;
	#endif	
	}

	if(strchr(flags,'R'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -R INTEGER," << setw(25) << "--repetitions=INTEGER";
	#else		
		cout << setw(20) << "    -R INTEGER,";
	#endif		
		cout << "The number of repetitions for every simulation run" << endl;
	}
	if(strchr(flags,'E'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -E INTEGER," << setw(25) << "--engines=INTEGER";
	#else
		cout << setw(20) << "    -E INTEGER";
	#endif			
		cout << "Specify the number of simulation engines to be used" << endl;	
	}
	if(strchr(flags,'P'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -P DOUBLE," << setw(25) << "--precision=DOUBLE";
	#else
		cout << setw(20) << "    -P DOUBLE";
	#endif
		cout << "Set the precision for simulation results" << endl;
	}
	if(strchr(flags,'C'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -C DOUBLE," << setw(25) << "--confidence=DOUBLE";
	#else
		cout << setw(20) << "    -C DOUBLE";
	#endif
		cout << "Set the confidence level for simulation results" << endl;
	}
	if(strchr(flags,'G'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -G ID," << setw(25) << "--random-generator=ID";
	#else
		cout << setw(20) << "    -G ID";
	#endif
		cout << "Select the generator for random numbers" << endl;
	#ifdef HAVE_GETOPT_LONG
		cout << setw(42) << "";
	#else
		cout << setw(21) << "";
	#endif
		cout << "(Available are CMRG (default) and MTwister)" << endl;
	}
	if(strchr(flags,'S'))
	{
	#ifdef HAVE_GETOPT_LONG
		cout << setw(16) << "    -S INTEGER," << setw(25) << "--random-stream=INTEGER";
	#else
		cout << setw(20) << "    -S INTEGER";
	#endif
		cout << "Specify the random number stream to be used" << endl;
	#ifdef HAVE_GETOPT_LONG
		cout << setw(42) << "";
	#else
		cout << setw(21) << "";
	#endif
		cout << "(Otherwise, a stream will be randomly selected)" << endl; 
	}

	cout << endl;
}

void CmdParser::split (const string& text, const string& separators, list<string>& words) const
{
	int n = text.length();
	int start, stop;

	start = text.find_first_not_of(separators);
	while ((start >= 0) && (start < n)) 
	{
		stop = text.find_first_of(separators, start);
		if ((stop < 0) || (stop > n)) stop = n;
		words.push_back(text.substr(start, stop - start));
		start = text.find_first_not_of(separators, stop+1);
	}
}

bool CmdParser::parseVariables(string arg) 
{
	list<string> tokens;
	string seperator = ";";
	string inner_seperator = "=";
	split(arg, seperator, tokens);

	for (std::list<string>::const_iterator it=tokens.begin(); it!=tokens.end(); it++) 
	{
		std::list<string> inner;
		split((*it), inner_seperator, inner);
		if (inner.size() != 2) 
		{
			std::cerr << "Invalid syntax" << std::endl;
			exit(1);
		}
		std::list<string>::const_iterator va = inner.begin();
		const std::string& var_name = (*va++);
		const std::string& value = (*va);
		addValue( var_name, ValueString(value) );
	}
	return true;
};





