#include <configparser.h>
#include <errhandler.h>
#include <utility/xmlhelper.h>
#include <macros.h>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>

using namespace std;
XERCES_CPP_NAMESPACE_USE

ConfigParser::ConfigParser()
{
    try 
	{	
		// Initialisierung von Xerces-C
		XMLPlatformUtils::Initialize();
	}
		
	catch (const XMLException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.getMessage());
		cout << "Failed to set up Xerces-C++ for parsing configuration file! " << endl;
		cout << "Message: " << message << endl;
		XMLString::release(&message);
		exit(1);
	}
}

bool ConfigParser::parseFile(const char* configfile)
{	        
	// Parser
	XercesDOMParser* schemaparser = new XercesDOMParser();

	try 
	{		
		// Parser initialisieren
		schemaparser->setErrorHandler(new ErrHandler());
		schemaparser->setValidationScheme(AbstractDOMParser::Val_Auto);
		schemaparser->setDoSchema(true);
		schemaparser->setDoNamespaces(true);
    
		// Parsen, Excetions registrieren
		schemaparser->parse(configfile);

		// Fehler?
		if(((ErrHandler*)schemaparser->getErrorHandler())->getErrCount())
		{
			// Abbruch
			return false;
		}
		
		// Konfiguration auslesen
		DOMNodeIterator* iterator = 
			createNodeIterator(schemaparser->getDocument()->getDocumentElement(),NULL,false);
		DOMElement* cfg = (DOMElement*)iterator->nextNode();
	
		while(cfg)
		{
			if(equals(cfg->getNodeName(), "parser"))
			{
				addValue(_XML_SCHEMA_, parseAttribute(cfg,"schema"));
				addValue(_XML_NAMESPACE_, parseAttribute(cfg,"namespace"));
				addValue(_XML_PROCESSING_, parseAttribute(cfg,"processing"));
				addValue(_XML_STOPONERROR_, parseAttribute(cfg,"exitOnFirstError"));
			}
			else if(equals(cfg->getNodeName(), "engine"))
			{
				addValue(_ENGINE_BINARY_, parseAttribute(cfg, "binary"));
				addValue(_ENGINES_, parseAttribute(cfg, "number"));
				addValue(_NICE_LEVEL_, parseAttribute(cfg, "niceLevel"));
			}
			else if(equals(cfg->getNodeName(), "prng"))
			{
				addValue(_PRNG_ID_, parseAttribute(cfg, "factoryId"));
				// optional
				if(cfg->hasAttribute(conv("streamId")))
					addValue(_PRNG_STREAM_, parseAttribute(cfg, "streamId"));
			}	
			else if(equals(cfg->getNodeName(), "simulation"))
			{
				addValue(_REPETITIONS_, parseAttribute(cfg, "repetitions"));
				addValue(_PRECISION_, parseAttribute(cfg, "precision"));
				addValue(_CONFIDENCE_, parseAttribute(cfg, "confidence"));
			}
			else if(equals(cfg->getNodeName(), "analysis"))
			{
				addValue(_ANALYSIS_ID_, parseAttribute(cfg, "factoryId"));
			}
			else if(equals(cfg->getNodeName(), "output"))
			{
				addValue(_SHOW_VERBOSE_, parseAttribute(cfg, "verbose"));
				addValue(_SHOW_PROGRESS_, parseAttribute(cfg, "progress"));
			}
			else if(equals(cfg->getNodeName(), "data"))
			{
				addValue(_DATA_FILE_, parseAttribute(cfg, "file"));
			}
			else if(equals(cfg->getNodeName(), "plot"))
			{
				addValue(_PLOT_FILE_, parseAttribute(cfg, "file"));
			}
			else if(equals(cfg->getNodeName(), "variable"))
			{
				//FIXME: parseAttribute kann hier Variable durch Wert ersetzen,
				//
				//addValue((string)parseAttribute(cfg,"name"), parseAttribute(cfg,"value"));
				addValue(conv(cfg->getAttribute(conv("name"))), parseAttribute(cfg,"value"));
			}
			cfg = (DOMElement*)iterator->nextNode();
		}
	}
	catch (const XMLException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.getMessage());
		cout << "Exception message is: " << endl << message << endl;
		XMLString::release(&message);
		return false;
	}
	catch (const DOMException& toCatch)
	{
		char* message = XMLString::transcode(toCatch.msg);
		cout << "Exception message is: "<< endl << message << endl;
		XMLString::release(&message);
		return false;
	} 
	catch (...)
	{
		cout << "Unexpected Exception" ;
		return false;
	}

	if (schemaparser) delete schemaparser;
	return true;
}

