#ifndef _OUTPUTINTERFACE_H_
#define _OUTPUTINTERFACE_H_

#include <stdexcept>
#include <vector>
#include <xercesc/dom/DOM.hpp>

XERCES_CPP_NAMESPACE_USE

//! Exception fuer Fehler bei Ausgabe.
class OutputException : public std::runtime_error
{
	public:
        OutputException(const std::string &msg) : std::runtime_error(msg){}
};


//! Ausgabe von Simulationsergebnissen.
/*!
 * Dieses Template stellt eine Basis fuer die Ausgabe von Ergebnissen dar. 
 * Es werden das simulierte Dokument, ein Stream zur Ausgabe und die 
 * Simulationsergebnisse gekapselt. Das Format (Typ) der Ergebnisse kann 
 * ueber den Typ-Parameter festgelegt werden.
 */
template<typename T>
class OutputInterface
{
	public:

		//! Konstruktor.
		/*!
		 * Es werden das Dokument und der Ausgabe-Stream initialisiert,
		 * sofern als Argumente uebergeben.
		 *
		 * @param doc DOM-Dokument der Simulationsbeschreibung.
		 * @param ostream Zu verwendender Ausgabe-Stream.
		 */
        OutputInterface(const DOMDocument* doc = NULL, std::ostream* ostream = NULL) 
			: domdocument(doc), out(ostream)
        {}

        //! Destruktor.
        virtual ~OutputInterface()
        {}

		//! Simulationsergebnisse.
		/*!
		 * Diese Klasse verwaltet die relevanten Ergebnisse einer
		 * Simulation. 
		 */
        class Result
        {
			public:

				template<typename C>
                Result(const C& structure)
                {
					values = structure.values;
				
					estimate = structure.estimate;
					variance = structure.variance;
					error = structure.error;
					delta = error*estimate;
					
					confidence = structure.confidence;
					precision = structure.precision;
                }

				//! Zugriff auf Anzahl der Messwerte.
				inline const T& Values() const
				{
					return values;
				}
				//! Zugriff auf bestimmte Abschaetzung.
                inline const T& Estimate() const
                {
					return estimate;
                };
				//! Zugriff auf Fehler der Abschaetzung.
                inline const T& Variance() const
                {
					return variance;
                };					
				//! Zugriff auf die bestimmte Varianz.
                inline const T& Error() const
                {
					return error;
                };
				//! Zugriff auf Haelfte des Konfidenzintervalls
				inline const T& Delta() const
				{
					return delta;
				};
				//! Zugriff auf das Konfidenz-Intervall.
                inline const T& Conf() const
                {
					return confidence;
                };
				//! Zugriff auf die Genauigkeit.
                inline const T& Prec() const
                {
					return precision;
                };

        private:

				//! Werte
				T values;
				
				//! Abschaetzung.
                T estimate;
				//! Varianz.
                T variance;
                //! Fehler
				T error;
				//! Haelfte des Konfidenzintervalls
				T delta;
				
				//! Konfidenz-Intervall.
                T confidence;
				//! Genauigkkeit.
                T precision;
        };

		//! Erzeugen einer Ausgabe.
		/*!
		 * Diese virtuelle Methode wird von allen abgeleiteten
		 * Klassen implementiert und erzeugt Ausgaben zu gegebenen
		 * Simulationsergebnissen. Ueber die Methode addResult
		 * muessen zuvor Ergebnisse bereitgestellt werden.
		 */
        virtual void generate() = 0;

		//! Hinzufuegen von Ergebnissen.
		/*!
		 * Mit dieser Methode koennen Ergebnisse, die zur Ausgabe
		 * bestimmt sind, hinzugefuegt werden.
		 * 
		 * @param result Ein Ergebnisdatum.
		 */
        const unsigned int addResult(Result result)
        {
			result_list.push_back(result);
			return result_list.size();
        }

		//! Anzahl der verfuegbaren Ergebnisse.
		/*!
		 * Diese Methode gibt die Groesse der internen Ergebnisliste
		 * wieder.
		 *
		 * @return Anzahl der aufgenommenen Ergebnisse.
		 */
        const unsigned int Size() const
        {
			return result_list.size();
        }

		//! Zugriff auf Ergebnis.
		/*!
		 * Diese Methode erlaubt den Zugriff auf ein Ergebnis der 
		 * Ergbnisliste. Der Index muss aus dem Intervall [0,Size()-1] 
		 * sein.
		 *
		 * @param index Ergebnisindex.
		 * @return Ergebnis.
		 */
        const Result& getResult(const unsigned int index) const
        {
			return result_list.at(index);
        }

		//! Zugriff auf XML-Dokument.
		/*!
		 * Das XML-Dokument einer Simulationsbeschreibung liegt in Form
		 * eines DOM-Dokuments vor und kann fuer weitere Informattionen 
		 * zur Simulation, deren Ergebnisse dargestellt werden, untersucht 
		 * werden.
		 * 
		 * @return DOM-Dokument zu Ergebnissen.
		 */
		const DOMDocument* getDocument() const
		{
			return domdocument;
		}
		
private:

		//! Ergebnisliste.
        std::vector<Result> result_list;
		//! DOM-Dokument.
        const DOMDocument* domdocument;

protected:
		
		//! Stream fuer Ausgabe.
        std::ostream* out;
};


#endif
