#include <sprouting.h>
#include <packet.h>
#include <targetbuffer.h>
#include <router.h>
#include <configuration.h>
#include <macros.h>

#include <stack>
#include <assert.h>

REGISTER_ELEMENT(SPRouting, "ShortestPath")

std::ostream& operator<<(std::ostream& target, std::vector<Distance> toShow);
std::ostream& operator<<(std::ostream& target, Matrix<Distance>& toShow);

//! Hilfstrukur zum Traversieren des Netzes.
struct BNodeFilter : public Node::Filter
{
	//! Nur Puffer sind zulaessig, keine Router oder Generatoren.
	bool valid(Node* node) const 
	{ 
	    return node->rtti()==Component::BUFFER
		    || node->rtti()==Component::TARGET; 
	}
};

bool SPRouting::init()
{
	const ComponentVector& router = Components(Component::ROUTER);
	const ComponentVector& buffers = Components(Component::BUFFER);
	const ComponentVector& targets = Components(Component::TARGET);

	Matrix<Distance> minima(buffers.size(), targets.size());
	std::stack<Buffer*> reached;
	Buffer* current = NULL;
	Node::Filter* filter = new BNodeFilter;
	int distance;
			
	// Netz ausgehend von den Targets expandieren und
	// minimale Weglaengen initialisieren
	for(unsigned int t=0; t<targets.size(); t++)
	{
		reached.push((Buffer*)targets[t]);

		// Tiefensuche ausgehend vom aktuellen Target, minimale
		// Weglaengen setzen und Vorgaenger weiter expandieren, 
		// falls neuer Wert kleiner als aktueller
		while(!reached.empty())
		{
			current = reached.top();
			reached.pop();

			if(current->rtti()==Component::TARGET)
				distance = 0;
			else
				distance = minima[current->Index()][t].value+1;
						
			// Neue minimale Weglaenge fuer Vorgaenger setzen, Vorgaenger 
			// dann auf den Stack
			const std::vector<Node*>& prev = current->prevNodes(filter);
			for(unsigned int i=0; i<prev.size(); i++)
			{
				// Abbruchkriterium fuer Zyklen: Ist die gesetzte
				// Weglaenge bereits kleiner, dann waren wir schon
				// hier!
				if(minima[prev[i]->Index()][t].value <= distance)
					continue;

				minima[prev[i]->Index()][t].value=distance;
				reached.push((Buffer*)prev[i]);
			}
		}
	}

	_DEBUG(minima);
	
	// fuer jeden Puffer eine Bitmaske
	bitmasks.resize(buffers.size());

	// Annahme: Matrix hat Zeile fuer jeden Puffer
	assert(buffers.size() == minima.getYSize());

	// Schleife ueber alle Router r
	for(unsigned int r=0; r<router.size(); r++)
	{
		const NodeVector& next = ((Node*)router[r])->nextNodes();

		// fuer alle Ziele t...
		for(unsigned int t=0; t<targets.size(); t++)
		{
			// ...die von Router r erreichbar
			if(!router[r]->getBitmask().test(t))
				continue;
			
			unsigned int shortest=UINT_MAX;
		
			// fuer alle Knoten an Ausgaengen...
			for(unsigned int n=0; n<next.size(); n++)
			{
				unsigned int local = UINT_MAX;
				
				// Knoten am Ausgang ist Puffer
				if(next[n]->rtti() == Component::BUFFER)
					local = minima[next[n]->Index()][t].value;
				
				// Knoten am Ausgang ist Target
				else if(next[n]->rtti() == Component::TARGET)
					local = next[n]->Index()==t ? 0 : local;

				else _ERROR("Expected buffer or target following router "
						<< router[r]->InternalName() << "!");
					
				if(local<shortest)
					shortest=local;
			}
			
			// Annahme: Ziel ueber mindestens einen Ausgang erreichbar.
			assert(shortest < UINT_MAX);
			
			// nochmal alle Ausgaenge
			for(unsigned int n=0; n<next.size(); n++)
			{
				// Targets ueberspringen
				if(next[n]->rtti() == Component::TARGET)
					continue;
				
				// wenn Weg zu Ziel t minimal...
				if(minima[next[n]->Index()][t].value == shortest)
				{
					// Annahme: Ziel ist auch erreichbar
					assert(next[n]->getBitmask().test(t));
					
					// ..dann Bit setzen 
					bitmasks[next[n]->Index()].set(t);
				}
			}
		}
	}
	
	delete filter;
	return true;
}

Bitmask* SPRouting::validOutput(const Packet* ptr, const Bitmask* remaining, const Wire* outwire)
{
	const Bitmask *valid = NULL;
	
	if(outwire->To()->rtti() == Component::TARGET)
		valid = &outwire->To()->getBitmask();
	
	else if(outwire->To()->rtti() == Component::BUFFER)
		valid = &bitmasks[outwire->To()->Index()];
		
	else _ERROR("Expected buffer or target here!");
		
	// Test gegen Bitmaske mit zulaessigen Zielen
	if((*remaining) <= (*valid))
	{
		// Annahme: Ueber Knoten sollte Ziel erreichbar sein
		assert(!valid->empty());
		
		return new Bitmask((*remaining) & (*valid));
	}
	return NULL;	
}

std::ostream& operator<<(std::ostream& target, std::vector<Distance> toShow)
{
	for(unsigned int i=0;i<toShow.size()-1;i++)
	{	
				
		if(toShow[i].value < INT_MAX) 
		target << "t" << i << ": " << toShow[i].value << " ";
	}	
	return target;
}

std::ostream& operator<<(std::ostream& target, Matrix<Distance>& toShow)
{
	std::cerr << "Shortest Paths (to reachable targets): " << std::endl;
	for(int i=0;i<toShow.getYSize()-1;i++)
	{
		std::cerr << Components(Component::BUFFER)[i]->toString() << " -> ";
		target << toShow[i] << std::endl;
	}
	return target;
}

