#include <wfrouting.h>
#include <configuration.h>
#include <targetbuffer.h>
#include <router.h>
#include <utility/coordinate.h>
#include <macros.h>

#include <cassert>

REGISTER_ELEMENT(WFRouting, "WestFirst")

bool WFRouting::init()
{
	const ComponentVector& router = Components(Component::ROUTER);
	const ComponentVector& buffers = Components(Component::BUFFER);
	const ComponentVector& targets = Components(Component::TARGET);
	
	// Jeweils eine Bitmaske fuer jeden Puffer
	_bitmasks_p1.resize(buffers.size());
	_bitmasks_p2.resize(buffers.size());

	// Alle Router
	for(unsigned int r=0; r<router.size(); r++)
	{	
		// Koordinaten des untersuchten Routers.
		std::vector<int>& currentpos = ((Node*)router[r])->position;
		
		// Annahme: Position des Routers vorhanden
		assert(currentpos.size() == 2);
					
		// Alle Ausgaenge eines Routers
		for(unsigned int o=0; o<((Node*)router[r])->Out().size(); o++)
		{
			// Nachfolger
			Node* next = ((Node*)router[r])->Out()[o]->To();
			
			if(next->rtti() == Component::TARGET)
				continue;
				
			if(next->rtti() != Component::BUFFER)
				_ERROR("Expected buffer or target following router " 
						<< router[r]->InternalName() << "!");

			// Koordinaten des Nachfolgers
			std::vector<int>& nextpos = next->position;
			
			// Annahme: Position des Knoten am Ausgang vorhanden
			assert(nextpos.size() == 2);
				
			//
			// Phase 1
			//
			for(unsigned int t=0; t<targets.size(); t++)
			{
				// Koordinaten des untersuchten Zieles.
				std::vector<int>& targetpos = ((Node*)targets[t])->position;

				// Annahme: Position des Ziels bekannt
				assert(targetpos.size() == 2);

				// Bedingung: Verbindung zu Ziel ueber anderen Gitterknoten 
				if(!next->getBitmask().test(t) || currentpos==targetpos)
					continue;

				// Ziel westlich - nur westlicher Ausgang zulaessig
				if(westOf(targetpos, currentpos)){
					if(westOf(nextpos, currentpos))
						_bitmasks_p1[next->Index()].set(t);
					continue;
				}

				// Ziel noerdlich - Ausgang im Norden und Westen zulaessig
				if(northOf(targetpos, currentpos)){
					if(northOf(nextpos, currentpos) || westOf(nextpos, currentpos))
						_bitmasks_p1[next->Index()].set(t);
				}
				// Ziel suedlich - Ausgang im Sueden und Westen zulaessig
				if(southOf(targetpos, currentpos)){
					if(southOf(nextpos, currentpos) || westOf(nextpos, currentpos))
						_bitmasks_p1[next->Index()].set(t);
				}
				// Ziel oestlich - Ausgang im Osten zulaessig
				if(eastOf(targetpos, currentpos)){
					if(eastOf(nextpos, currentpos))
						_bitmasks_p1[next->Index()].set(t);
				}
			}

			//
			// Phase 2
			//
			for(unsigned int t=0; t<targets.size(); t++)
			{
				// Koordinaten des untersuchten Zieles
				std::vector<int>& targetpos = ((Node*)targets[t])->position;

				// Annahme: Position des Ziels bekannt
				assert(targetpos.size() == 2);

				// Bedingung: Verbindung zu Ziel in Osten, Sueden oder Norder ueber 
				// anderen Gitterknoten 
				if(westOf(targetpos, currentpos) || currentpos==targetpos
					|| !next->getBitmask().test(t))
					continue;

				// Ziel noerdlich - Ausgang im Norden zulaessig
				if(northOf(targetpos, currentpos)){
					if(northOf(nextpos, currentpos))
						_bitmasks_p2[next->Index()].set(t);
				}
				// Ziel suedlich - Ausgang im Sueden zulaessig
				if(southOf(targetpos, currentpos)){
					if(southOf(nextpos, currentpos))
						_bitmasks_p2[next->Index()].set(t);
				}
				// Ziel oestlich - Ausgang im Osten zulaessig
				if(eastOf(targetpos, currentpos)){
					if(eastOf(nextpos, currentpos))
						_bitmasks_p2[next->Index()].set(t);
				}
			}
		}
	}
	return true;
}

Bitmask* WFRouting::validOutput(const Packet* ptr, const Bitmask* remaining, const Wire* outwire)
{
	// alle Quellen
	const ComponentVector& sources = Components(Component::SOURCE);	
	
	// Ueber Ausgang erreichbare und zulaessige Ziele
	const Bitmask *valid = NULL;

	std::vector<int>& srcpos = ((Node*)sources[ptr->sourcenr])->position;
	std::vector<int>& nowpos = ptr->front()->Location().where->position;

	if(outwire->To()->rtti() == Component::TARGET)
		valid = &outwire->To()->getBitmask();
	
	else if(outwire->To()->rtti() == Component::BUFFER)
	{
		const NodeVector& prevs = ptr->front()->Location().where->prevNodes();

		// kein direktes Umdrehen erlauben
		if(prevs.size() && prevs.front()->position == outwire->To()->position) return NULL;

		// Phase 2
		if(northOf(nowpos, srcpos) || southOf(nowpos, srcpos) || eastOf(nowpos, srcpos))
			valid = &_bitmasks_p2[outwire->To()->Index()];
		// Phase 1
		else
			valid = &_bitmasks_p1[outwire->To()->Index()];
	}

	// Test der Bitmaske mit offenen Zielen gegen Bitmaske mit zulaessigen Zielen
	if((*remaining) <= (*valid))
	{
		// Annahme: Ueber Knoten sollte Ziel erreichbar sein
		assert(!valid->empty());
		
		return new Bitmask((*valid) & (*remaining));
	}
	return NULL;
}

