#include <sequencer.h>
#include <xmlsequencer.h>
#include <simulation.h>
#include <environment.h>
#include <macros.h>

#include <output/text.h>
#include <output/gnuplot.h>

#include <iomanip>
#include <csignal>
#include <cassert>

Sequencer* Sequencer::_instance = new XmlSequencer;


Sequencer::~Sequencer()
{
	if(_results)
	{
		for(unsigned int i=0; i<_setup.nofsim; i++)
		{
			delete _results[i];
		}

		delete [] _results;
	}

	_parameters.clear();
}


void Sequencer::SignalHandler(int signr) 
{
	// Aktion fuer ctrl-c (SIGINT)
	static enum
	{
		PRINT,	// Ergebnisse ausgeben 
		ABORT	// Beenden
	} 
	action = PRINT;
			
	switch(signr)
	{	
		
		//
		// Ergebnisse/Abbruch ueber ctrl-c
		// 
		case SIGINT:
				
				if(action == PRINT)
				{
					//signal(SIGINT, SIG_IGN);
					signal(SIGALRM, Sequencer::SignalHandler);

					action = ABORT;

					//
					// Timer fuer wiederholtes ctrl-c (2 Sekunden) 
					//
					alarm(2);
				}
				else 
				{
					if(Env()->isSet(_DATA_FILE_) || Env()->isSet(_PLOT_FILE_))
						_instance->output();
					
					exit(0);
				}
		
		// 
		// Ergebnisse ueber 'kill -12 PID'
		// 
		case SIGUSR2:				
		
				// Daten aktualiesieren
				if(!_instance->_repetition)
				{
					for(unsigned int i=0; i<Analysis::Declared(); i++)
					{
						_instance->_results[_instance->_simulation][i] 
							= Analysis::CurrentResult(i+1);
					}
				}
				
				if(Analysis::Declared()) _instance->intermediate();	
				
				signal(signr, Sequencer::SignalHandler);

				break;
	
		//		
		// Timer fuer Abbruch mit ctrl-c abgelaufen
		// 
		case SIGALRM:

				signal(SIGALRM, SIG_DFL);

				assert(action == ABORT);

				action = PRINT;	

				break;
				
		default:
				
			_ERROR("Unexpected signal caught!");
	}
}


void Sequencer::load()
{
	XmlLoad::load();

	for(unsigned int i=0; i < _parameters.size(); i++)
	{
		Environment::setParameter(_parameters[i].name, _parameters[i].value);

		_parameters[i].value += _parameters[i].increment;
	}
	
	Simulation::Load();
}


void Sequencer::start()
{
	unsigned int i=0;
	unsigned int rep = Env()->getValue(_REPETITIONS_);
	
	assert(_setup.nofobs == Analysis::Declared());
	
	// Datenstruktur fuer Ergebnisse
	_results = new Analysis::Result*[_setup.nofsim];
	for(; i<_setup.nofsim; i++)
	{
		_results[i] = new Analysis::Result[Analysis::Declared()];
	}
	
	// Simulationssequenz
	for(_simulation=0;;)
	{
		// Simulation wiederholen
		for(_repetition=0; _repetition<rep; _repetition++)
		{
			Simulation::Start();
		
			for(i=0; i<Analysis::Declared(); i++)
			{
				const Analysis::Result& current = Analysis::CurrentResult(i+1);
				Analysis::Result& old = _results[_simulation][i];
				
				if(!_repetition || old.values < current.values)
				{
					old = current;
				}
			}
					
			Analysis::Reset();
		}	

		if(_simulation + 1 < _setup.nofsim)
		{
			// Parameter aktualisieren
			for(unsigned int i=0; i < _parameters.size(); i++)
			{
				Environment::setParameter(_parameters[i].name, _parameters[i].value);

				_parameters[i].value += _parameters[i].increment;
			}

			// Simulation neu laden
			Simulation::Reload();

			_simulation++;
			
			continue;
		}
		
		break;
	}

	output();
}

void Sequencer::output() const
{
	std::string data_file = Env()->getValue(_DATA_FILE_);
	std::string plot_file = Env()->getValue(_PLOT_FILE_);
	
	ofstream* fout = NULL;
	char info[128];

	if(plot_file != snull)
	{
		sprintf(info,"*** Generating gnuplot output through C++ interface "); 	
		std::cout << std::endl << left << setfill('*') << setw(100) << info 
			<< setfill(' ') << std::endl;

		GnuplotOutput gpl(_document, plot_file, &std::cout);
	
		for (unsigned int s=0; s<=_simulation; s++)
			for (unsigned int i=0; i<_setup.nofobs; i++)
				gpl.addResult(OutputInterface<double>::Result(_results[s][i]));
		
		try 
		{
			gpl.generate();
		} 
		catch (OutputException oe) 
		{
			cout << "++ Error: " << oe.what() << std::endl;
		}
	}

	sprintf(info, "*** Generating textual output for %d result(s) ", 
			_setup.nofobs * (_simulation+1)); 
	
	std::cout << std::endl << left << setfill('*') << setw(100) << info
		<< setfill(' ') << std::endl;
	
	if(data_file != snull)
	{
		fout = new ofstream(data_file.c_str());

		if(!fout->is_open())
		{
			_WARNING("Failed to open file " << data_file
				<< "! Standard output will be used instead.");
			fout = NULL;
		}
	}
	
	TextOutput text_output(_document, fout ? fout : &std::cout);
	for (unsigned int s=0; s<=_simulation; s++)
		for (unsigned int i=0; i<_setup.nofobs; i++)
			text_output.addResult(OutputInterface<double>::Result(_results[s][i]));

	text_output.generate();

	if(fout)
		std::cout << "++ Results successfully written to " << data_file << std::endl;
	
	std::cout << std::endl << left << setw(100) << setfill('*') << "*** Finished " 
		<< setfill(' ') << std::endl;
}


void Sequencer::intermediate() const
{
	char info[128];

	sprintf(info, "*** Generating textual output for %d result(s) ", 
			_setup.nofobs * (_simulation+1));	
	
	std::cout << std::endl << left << setfill('*') << setw(100) << info
		<< setfill(' ') << std::endl;
	
	TextOutput text_output(_document, &std::cout);
	for (unsigned int s=0; s<=_simulation; s++)
		for (unsigned int i=0; i<_setup.nofobs; i++)
			text_output.addResult(OutputInterface<double>::Result(_results[s][i]));

	text_output.generate();

	std::cout << std::endl << left << setw(100) << setfill('*') << "*** Finished " 
		<< setfill(' ') << std::endl;
}

