#include <sourcebuffer.h>
#include <bufferqueue.h>
#include <targetbuffer.h>
#include <distribution.h>
#include <traffic.h>
#include <packetforwarding.h>
#include <configuration.h>
#include <switching.h>
#include <packet.h>
#include <macros.h>

#include <utility/xmlhelper.h>
#include <utility/bitmask.h>

#include <cassert>

using namespace std;

#ifdef DEBUG
vector<unsigned int> counter;
#endif

REGISTER_ELEMENT(SourceBuffer, "sourcebuffer")


SourceBuffer::SourceBuffer(DOMElement* scope) : Buffer(scope) 
{
	deadline = 0;
	constPacketSize = 0;
	maxPriority = 0;

	mc_dist = NULL;
	lc_dist = NULL;
	pr_dist = NULL;
	tr_dist = NULL;

	p_accepted = false;
}

bool SourceBuffer::init(const DOMElement* scope) 
{
	if(scope)
		setScope(scope);
	assert(Scope());

	constPacketSize = getAttribute("packetSize");
	
	//
	// Alle Switching-Verfahren ausser Wormhole erfordern eine garantierte
	// Puffergroesse >= der Paketgroesse.
	//
	if((((string)parseAttribute(Configuration::Instance()->Settings(),"switchingType")) 
		!= "Wormhole") && (constPacketSize > (uint)getAttribute("minSpace")))
		_ERROR("Selected packet size of " 
				<< constPacketSize 
				<< " exceeds guaranteed buffer space of " 
				<< (uint)getAttribute("minSpace")
				<< "!")

	deadline = getAttribute("deadline");
	maxPriority = getAttribute("maxPriority");

	// Die Queue
	if(!BQ) BQ = new BufferQueue(UINT_MAX);
	
	// Multicast
	if(!mc_dist)
	{
		mc_dist = (Distribution*)FactoryElement::Factory::self().
			create((string)getAttribute("multicast"));
	
		assert(mc_dist);
	}
		
	// Lokaler Verkehr
	if(!lc_dist)
	{
		lc_dist = (Distribution*)FactoryElement::Factory::self().
			create((string)getAttribute("spatialDistribution"));

		assert(lc_dist);
	}
	
	// Prioritaeten
	if(!pr_dist)
	{
		pr_dist = (Distribution*)FactoryElement::Factory::self().
			create((string)getAttribute("priorityDistribution"));

		assert(pr_dist);
		
		// Verteilung der Prioitaeten initialisieren
		pr_dist->setDistElements(maxPriority+1);
		pr_dist->setAttributePrefix("pd");
		for(unsigned int i=0; i<=maxPriority; i++)
		{
			stringstream pr_key;
			pr_key << i;
			pr_dist->indexMapping()[pr_key.str()] = i;
		}
	
		pr_dist->init(Scope());
	}	
		
	// Verkehr
	string traffic = getAttribute("traffic");
	if(!tr_dist)
	{
		tr_dist = (TrafficDistribution*)FactoryElement::Factory::self().create(traffic);
	}
		
	assert(tr_dist);
	
	tr_dist->init(getFirstChildByName(Scope(), traffic.c_str(), equalsIgnoreCase));

	return true;
}

double SourceBuffer::Measure(TypeInfo type)
{
	switch (type) 
	{
	//case ePrioriSThroughput:
	case eSThroughput:
		return BQ->Popped() ? 1.0 : 0.0;
		break;
	//case eLoad:
	//	return (hadPacket);
	//	break;
	default:
		_ERROR("Unexpected measure type: " << type);
	}
	return 0.0;
}

void SourceBuffer::takeMeasure(const Description& descr)
{	
	switch (descr.type) 
	{
	// Durchsatz an Flits
	case eSThroughput:
		Analysable::takeMeasure(descr);
	default:
		break;
	}
}
	
int SourceBuffer::putPacketOnLine(Wire* wire, bool& head)
{
	// altes, nicht ins Netz gelangtes Paket entfernen
	if(BQ->spaceInUse() && BQ->front()->head() && !p_accepted)
	{		
		_PACKETMOVE(toString() << " dropped:", BQ->front()->Owner()->toString(), 
				"Queue: " << BQ->freeSpace() << " of " << BQ->queueSpace());
		
		BQ->clear();	
	}
	
	// Versuch neues Paket zu generieren
	if( tr_dist->ifNowGenerate( Tick() ) && BQ->empty() && State()==Simulation::NORMAL)
		generatePacket();
		
	return Buffer::putPacketOnLine(wire,head);
}

void SourceBuffer::reset()
{	
	// Queue leeren, Register zuruecksetzen
	BQ->clear();
	// Paket-Verteilung zuruecksetzen
	tr_dist->reset();
}

void SourceBuffer::checkReachableTargets() 
{
	const ComponentVector& tvector = Components(Component::TARGET);
		
	assert(tvector.size());

	// Bitmaske erstellen
	generateBitmask();
	
    for (unsigned int target = 0; target<tvector.size(); target++ )
	{
		if ( getBitmask().test(target))
		{
			stringstream mc_key;

			reachableTargets.push_back(target);
			mc_key << reachableTargets.size();

			lc_dist->indexMapping()[tvector[target]->InternalName()] 
				= reachableTargets.size()-1;

			mc_dist->indexMapping()[mc_key.str()] = reachableTargets.size()-1;
		}
	}
	
	// initialize the multicast and local distribution
	// this one sets the possible multicast size
	assert(mc_dist && lc_dist);
	mc_dist->setDistElements(reachableTargets.size());
	mc_dist->setAttributePrefix("mc");
	mc_dist->init(Scope());
	
	lc_dist->setDistElements(reachableTargets.size());
	lc_dist->setAttributePrefix("sd");
	lc_dist->init(Scope());
	
	_INFO(this->toString() << " multicast probabilities:\n\t\t" << mc_dist);
	_INFO(this->toString() << " target selection probabilities:\n\t\t" << lc_dist);
	_INFO(this->toString() << " priority selection probabilities:\n\t\t" << pr_dist);

#ifdef DEBUG
	// dieser vektor soll zaehlen, wie oft welches ziel fuer ein paket bestimmt worden ist
	for (unsigned int i=0; i<reachableTargets.size(); i++)
		counter.push_back(0);
#endif
	
	// TODO: Sinnvoller Test fuer Kombination Multicast/lokaler Verkehr
	//_ERROR("The combination of multicast and spatial distribution does not make sense.");
}

bool SourceBuffer::generatePacket()
{
	// Annahme: Es existieren erreichbare Ziele!
	assert(reachableTargets.size()>0);
	
	// Paket der Groesse 0 generieren
	Packet* ptr = Coroutine()->switching()->createPacket(constPacketSize);	
	
	// Paket initialisieren
	ptr->sourcenr = Index();
	ptr->createTick = Tick();
	ptr->priority = maxPriority ? pr_dist->choose() : 0;
	ptr->deadline = deadline+Tick();

	// Multicast Anzahl ziehen
	unsigned int mc_nr = mc_dist->choose()+1; // ohne 0

	// TODO: Vorsicht baustelle
	for (unsigned int i=0; i<mc_nr; i++) 
	{
		unsigned int choosen = lc_dist->choose();
		
		// Distribution darf kein Ziel zweimal liefern
		assert(!ptr->getBitmask().test(reachableTargets[choosen]));
		
		ptr->setTarget(reachableTargets[choosen]);
#ifdef DEBUG
		counter[choosen]++;
#endif
		// letzt gezogenes element entfernen
		lc_dist->remove_last();		
	}

#ifdef DEBUG
	if (Tick()%100) {
		unsigned int sum = 0;
		for (unsigned int i=0; i<counter.size(); i++) {
			sum += counter[i];
		}

		std::cerr <<  Name() << " " << StrX(Scope()->getAttribute(conv("index"))) << " : " ;
		
		for (unsigned int i=0; i<counter.size(); i++)
			std::cerr << (double)counter[i]/(double)sum << " ";
		std::cerr << std::endl;
	}
#endif
	// fuer den naechsten takt alles zuruecksetzen, 
	// alle elemente koennen wieder gezogen werden
	lc_dist->reset();
	
	// Paket am Generator einpuffern
	FlitIterator it = ptr->begin();
	while(it!=ptr->end())
	{
		BQ->push(*it);
		(*it)->buffered(this, -1);
		it++;	
	}
	
	// Flag fuer ins Netz gelangtes Paket zuruecksetzen
	p_accepted = false;
	
	assert(BQ->front()->head());
	
	_PACKETMOVE(toString() << " generated:",ptr->toString(),
			"Queue: " << BQ->spaceInUse() << " of " << BQ->queueSpace());
	
	return true;
}

void SourceBuffer::flitTransmitted(Flit* flit, bool complete)
{
	// mindestens ein Empfaenger
	p_accepted = true;
	// vollstaendig vermittelt
	if(complete) BQ->pop();
}
