#include <test_registercontainer.h>

#include <switching/sfwpacket.h>
#include <utility/bitmask.h>
#include <packet.h> 
#include <flit.h>
#include <registercontainer.h>
#include <bufferqueue.h>

#include <buffer.h>
#include <router.h>

#include <simulation.h> // ticks

#include <cassert>

CPPUNIT_TEST_SUITE_REGISTRATION( Test_RegisterContainer );

using namespace std;

#define length 4 // min=2!
#define SET_NULL const_test_flit=NULL; test_flit=NULL; test_packet=NULL;
#define TICK Signature::InvalidateAll();

void Test_RegisterContainer::setUp()
{
	//
	// Hinweis: Normalerweise sind Komponenten Teil einer Konfiguration 
	// und diese Teil einer Simulation. Programmfragmente, die auf die
	// Komponenten der aktiven Konfiguration zugreifen wollen, erzeugen
	// mit diesem einfachen Setup Speicherzugriffsfehler.
	//

	Bitmask mask;
	
	// Buffer0 - Sender
	buffer0=new Buffer;
	((Buffer*)buffer0)->BQ=new BufferQueue(2*length);
	buffer0->setIndex(0);

	// Buffer1 - Empfaenger
	buffer1=new Buffer;
	((Buffer*)buffer1)->BQ=new BufferQueue(length);
	buffer1->setIndex(0);
	mask.clear();
	mask.set(3);
	mask.set(5);
	mask.set(17);
	buffer1->_bitmask = mask; // targets 3, 5, 17
	
	// Buffer2 - Empfaenger
	buffer2=new Buffer;
	((Buffer*)buffer2)->BQ=new BufferQueue(length);
	buffer2->setIndex(1);
	mask.clear();
	mask.set(1);	
	mask.set(35);
	buffer2->_bitmask = mask; // targets 1, 35
	
	// Router0 - Schnittstelle
	router0 = new Router;
	router0->setIndex(0);

	// Wire0
	wire0 = new Wire;
	wire0->setFrom(buffer0);
	wire0->setTo(router0);
	wire0->setOutIndex(buffer0->Out().size());
	buffer0->addOutWire(wire0);
	wire0->setInIndex(router0->In().size());
	router0->addInWire(wire0);

	// Wire1
	wire1 = new Wire;
	wire1->setFrom(router0);
	wire1->setTo(buffer1);
	wire1->setOutIndex(router0->Out().size());
	router0->addOutWire(wire1);
	wire1->setInIndex(buffer1->In().size());
	buffer1->addInWire(wire1);
	
	// Wire2
	wire2 = new Wire;
	wire2->setFrom(router0);
	wire2->setTo(buffer2);
	wire2->setOutIndex(router0->Out().size());
	router0->addOutWire(wire2);
	wire2->setInIndex(buffer2->In().size());
	buffer2->addInWire(wire2);

	// test packet 1
	packet1 = new SFWPacket(length);
	packet1->setTarget(1);
	packet1->setTarget(3);
	packet1->setTarget(5);
	packet1->number = 1;
	packet1->createTick = 0;
	packet1->sourcenr = 0;
	packet1->priority = 0;
	packet1->deadline = 0;

	assert(packet1->size() == length);

	// test packet 2
	packet2 = new SFWPacket(length);
	packet2->setTarget(1);
	packet2->setTarget(3);
	packet2->setTarget(5);
	packet2->number = 2;
	packet2->createTick = 0;
	packet2->sourcenr = 0;
	packet2->priority = 0;
	packet2->deadline = 0;

	assert(packet2->size() == length);

	container0 = ((Buffer*)buffer0)->BQ->Register();
	container1 = ((Buffer*)buffer1)->BQ->Register();
	container2 = ((Buffer*)buffer2)->BQ->Register();
	
	// Paket 1 in Puffer 0 einpuffern
	for (unsigned int i=0; i<packet1->size(); i++) 
	{
		((Buffer*)buffer0)->flitReceived(packet1->at(i));
	}		
	
	// Paket 2 in Puffer 0 einpuffern
	for (unsigned int i=0; i<packet2->size(); i++) 
	{
		((Buffer*)buffer0)->flitReceived(packet2->at(i));
	}
}

void Test_RegisterContainer::tearDown()
{
	delete buffer0;
	delete buffer1;
	delete buffer2;
	delete router0;
	delete wire0;
	delete wire1;
	delete wire2;
	//delete packet1;
	//delete packet2;
}

void Test_RegisterContainer::registration_test()
{
	// Testmasken: (3,5) und (1)
	Bitmask test1, test2;
	test1.set(3);
	test1.set(5);
	test2.set(1);
	
	// Ohne Registrierung gibt es kein Flit fuer Puffer 1 und 2
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer1)==NULL );
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer2)==NULL );

	// Puffer 1 fuer Ziele 3 und 5, Puffer 2 fuer Ziel 1 registrieren
	Bitmask* targets1 = new Bitmask;
	targets1->set(3);
	targets1->set(5);
	container0->registerCustomer(buffer1, targets1);
	Bitmask* targets2 = new Bitmask;
	targets2->set(1);
	container0->registerCustomer(buffer2, targets2);

	// Nach Registrierung Flits fuer Puffer 1 und 2 verfuegbar
	const_test_flit = container0->fetchableFlit(buffer1);
	CPPUNIT_ASSERT( const_test_flit && const_test_flit->Owner()->number == 1);
	const_test_flit = container0->fetchableFlit(buffer2);
	CPPUNIT_ASSERT( const_test_flit && const_test_flit->Owner()->number == 1);

	// Registrierte Bitmasken korrekt
	CPPUNIT_ASSERT( container0->validTargets(buffer1) && (*container0->validTargets(buffer1))==test1 );
	CPPUNIT_ASSERT( container0->validTargets(buffer2) && (*container0->validTargets(buffer2))==test2 );

	SET_NULL;
	
	// Bitmaske des erhaltenen Pakets fuer Puffer 1 korrekt
	test_flit = container0->fetchFlit(buffer1);
	CPPUNIT_ASSERT( test_flit && test_flit->Owner()->getBitmask()==test1 );

	// Registrierung von Puffer 2 aufheben, kein Flit mehr verfuegbar
	container0->unregisterCustomer(buffer2);
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer2) == NULL );
}


void Test_RegisterContainer::storing_test()
{
	// Initial kein Paket in Register-Container vorhanden
	CPPUNIT_ASSERT( container1->firstFlit()==NULL && container1->getPacket()==NULL );

	// Head von Paket 1 registrieren
	container1->announceFlit( packet1->at(0) );
	const_test_flit = container1->firstFlit();
	test_packet = container1->getPacket();
	CPPUNIT_ASSERT( const_test_flit && test_packet );
	CPPUNIT_ASSERT( const_test_flit->Owner() == test_packet && packet1 == test_packet );

	SET_NULL;
	
	// Rest von Paket 1 registrieren; vorderstes Flit/verfuegbares Paket unveraendert
	for( unsigned int i=1; i<packet1->size(); i++)
	{
		container1->announceFlit( packet1->at(i) );
	}
	const_test_flit = container1->firstFlit();
	test_packet = container1->getPacket();
	CPPUNIT_ASSERT( const_test_flit && test_packet );
	CPPUNIT_ASSERT( const_test_flit->Owner()==test_packet && packet1==test_packet );
	CPPUNIT_ASSERT( const_test_flit->Offset()==0 );
	
	SET_NULL;

	// Paket 2 registrieren; vorderstes Flit/verfuegbares Paket unveraendert
	for( unsigned int i=0; i<packet2->size(); i++)
	{
		container1->announceFlit( packet2->at(i) );
	}
	const_test_flit = container1->firstFlit();
	test_packet = container1->getPacket();
	CPPUNIT_ASSERT( const_test_flit && test_packet );
	CPPUNIT_ASSERT( const_test_flit->Owner()==test_packet && packet1==test_packet );	
	CPPUNIT_ASSERT( const_test_flit->Offset()==0 );

	SET_NULL;
	
	// Head von Paket 1 entfernen; Zugriff auf naechstes Flit moeglich
	CPPUNIT_ASSERT( packet1->at(0)==container1->firstFlit() );
	container1->eraseFlit( packet1->at(0) );
	const_test_flit = container1->firstFlit();
	test_packet = container1->getPacket();
	CPPUNIT_ASSERT( const_test_flit && test_packet );
	CPPUNIT_ASSERT( const_test_flit->Owner()==test_packet && packet1==test_packet );	
	CPPUNIT_ASSERT( const_test_flit->Offset()==1 );	
	
	SET_NULL;
	
	// Rest von Paket 1 entfernen; Zugriff auf Paket 2 moeglich
	for( unsigned int i=1; i<packet1->size(); i++)
	{
		CPPUNIT_ASSERT( packet1->at(i)==container1->firstFlit() );
		container1->eraseFlit( packet1->at(i) );
	}
	const_test_flit = container1->firstFlit();
	test_packet = container1->getPacket();
	CPPUNIT_ASSERT( const_test_flit && test_packet );
	CPPUNIT_ASSERT( const_test_flit->Owner()==test_packet && packet2==test_packet );
	CPPUNIT_ASSERT( const_test_flit->Offset()==0 );
	
	SET_NULL;
	
	// Paket 2 entfernen; Register leer
	for( unsigned int i=0; i<packet2->size(); i++)
	{
		CPPUNIT_ASSERT( packet2->at(i)==container1->firstFlit() );
		container1->eraseFlit( packet2->at(i) );
	}
	CPPUNIT_ASSERT( container1->firstFlit()==NULL && container1->getPacket()==NULL );
}


void Test_RegisterContainer::receiving_test()
{
	Bitmask *targets1, *targets2;
	SET_NULL;
	
	// Paket 1 in Register des Puffers 0
	const_test_flit = container0->firstFlit();
	CPPUNIT_ASSERT( const_test_flit && const_test_flit->Owner()->number == 1);
	test_packet = container0->getPacket();
	CPPUNIT_ASSERT( test_packet && test_packet->number == 1);
	
	SET_NULL;
	
	// Puffer 1 fuer Ziele 3 und 5, Puffer 2 fuer Ziel 1 registrieren
	targets1 = new Bitmask;
	targets1->set(3);
	targets1->set(5);
	container0->registerCustomer(buffer1, targets1);
	targets2 = new Bitmask;
	targets2->set(1);
	container0->registerCustomer(buffer2, targets2);

	// Puffer 1 empfaengt in Takt 0 Flit 0 von Paket 1 zu Zielen 3 und 5;
	// Anschliessend kein weiteres Flit verfuegbar
	test_flit = container0->fetchFlit(buffer1);
	CPPUNIT_ASSERT( test_flit && test_flit->Owner()->number==1 && test_flit->Offset()==0);
	CPPUNIT_ASSERT( test_flit->Owner() != container0->getPacket() );
	CPPUNIT_ASSERT( test_flit->Owner()->getBitmask().test(3) );
	CPPUNIT_ASSERT( test_flit->Owner()->getBitmask().test(5) );
	CPPUNIT_ASSERT( !test_flit->Owner()->getBitmask().test(1) );
	CPPUNIT_ASSERT( !test_flit->Owner()->getBitmask().test(17) );
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer1) == NULL );
	
	// Puffer 2 ist in Takt 0 verhindert; Anschliessend kein weiteres Flit verfuegbar
	container0->rejectFlit(buffer2);
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer2) == NULL );
	CPPUNIT_ASSERT( container0->finalState() );
	
	SET_NULL; TICK;

	// Puffer 1 und 2 empfangen Flits bis Puffer 1 fertig
	for(unsigned int i=1; i<packet1->size(); i++)
	{
		unsigned int oldsize = ((Buffer*)buffer0)->BQ->spaceInUse();
		
		CPPUNIT_ASSERT( container0->fetchableFlit(buffer1) );
		test_flit = container0->fetchFlit(buffer1);
		CPPUNIT_ASSERT( test_flit && test_flit->Owner()->number==1 && test_flit->Offset()==i);

		CPPUNIT_ASSERT ( container0->fetchableFlit(buffer2) );
		test_flit = container0->fetchFlit(buffer2);
		CPPUNIT_ASSERT( test_flit && test_flit->Owner()->number==1 && test_flit->Offset()==i-1);

		CPPUNIT_ASSERT( container0->finalState() );
		CPPUNIT_ASSERT( ((Buffer*)buffer0)->BQ->spaceInUse() == oldsize-1);
		
		SET_NULL; TICK;
	}
	
	// Queue des Puffers 0 enthaelt Paket 2 und Tail von Paket 1
	CPPUNIT_ASSERT( ((Buffer*)buffer0)->BQ->spaceInUse() == packet2->size()+1 );
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer1) == NULL);
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer2) != NULL);
	
	// Puffer 2 empfaengt Tail von Paket 1 zu Ziel 1; Paket 1 fertig; Paket 2 verfuegbar
	test_flit = container0->fetchFlit(buffer2);
	CPPUNIT_ASSERT( test_flit && test_flit->Owner()->number==1 && test_flit->Offset()==length-1 );
	CPPUNIT_ASSERT( test_flit->Owner()->getBitmask().test(1));
	CPPUNIT_ASSERT( !test_flit->Owner()->getBitmask().test(3));
	CPPUNIT_ASSERT( !test_flit->Owner()->getBitmask().test(5));
	CPPUNIT_ASSERT( !test_flit->Owner()->getBitmask().test(35));		
	CPPUNIT_ASSERT( container0->finalState() );
	test_packet = container0->getPacket();
	CPPUNIT_ASSERT( test_packet && test_packet->number==2 );

	SET_NULL; TICK;

	// Registrierungen nach Vermittlung von Paket 1 aufgehoben
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer1) == NULL);
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer2) == NULL);

	// Puffer 1 fuer Ziel 3, Puffer 2 fuer Ziel 1 registrieren
	targets1 = new Bitmask;
	targets1->set(3);
	container0->registerCustomer(buffer1, targets1);
	targets2 = new Bitmask;
	targets2->set(1);
	container0->registerCustomer(buffer2, targets2);

	// Puffer 1 und 2 empfangen alle Flits von Paket 2; durch fehlendes Ziel
	// 5 bleibt Paket 2 aber vollstaendig eingepuffert
	for(unsigned int i=0; i<packet2->size(); i++)
	{
		unsigned int oldsize = ((Buffer*)buffer0)->BQ->spaceInUse();
		
		CPPUNIT_ASSERT( container0->fetchableFlit(buffer1) );
		test_flit = container0->fetchFlit(buffer1);
		CPPUNIT_ASSERT( test_flit && test_flit->Owner()->number==2 && test_flit->Offset()==i);
		
		CPPUNIT_ASSERT ( container0->fetchableFlit(buffer2) );
		test_flit = container0->fetchFlit(buffer2);
		CPPUNIT_ASSERT( test_flit && test_flit->Owner()->number==2 && test_flit->Offset()==i);

		CPPUNIT_ASSERT( container0->finalState() );
		CPPUNIT_ASSERT( ((Buffer*)buffer0)->BQ->spaceInUse()==oldsize);
		
		SET_NULL; TICK;
	}
		
	// Registrierungen nach Empfang von Paket 2 aufgehoben
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer1) == NULL);
	CPPUNIT_ASSERT( container0->fetchableFlit(buffer2) == NULL);
	
	// Vorderstes Flit ist immer noch Flit 0 von Paket 2
	const_test_flit = container0->firstFlit();
	CPPUNIT_ASSERT( const_test_flit && const_test_flit->Owner()->number==2 && const_test_flit->Offset()==0);
	test_packet = container0->getPacket();
	CPPUNIT_ASSERT( test_packet && test_packet->number==2);
}

