#ifndef __PARETO_TRAFFIC__
#define __PARETO_TRAFFIC__

#include <traffic.h>

//! Selbstaehnlicher Verkehr ueber Pareto-Verteilung. 
/*!
 * Diese Paket-Verteilung von beschreibt selbstaehnlichen Verkehr durch
 * Verwendung einer Pareto-Verteilung.
 * 
 * @author Li Zhou
 */
class ParetoDistribution : public TrafficDistribution
{
public:
    
	//! Konstruktor.
	/*!
	 * Der Konstruktor initialisiert alle Klassenattribute der Basisklassen.
	 *
	 * @see TrafficDistribution
	 * @see XMLElement
	 * 
	 * @param psize Groesse der zu generierenden Pakete.
	 * @param scope Gueltigkeitsbereich innerhalb des XML-Dokumentes.
	 */
	ParetoDistribution(unsigned int psize=1, const DOMElement* scope=NULL)
		: TrafficDistribution(psize,scope) {}
	
	//! Implementierung der Basisklasse XMLElement.
	/*!
	 * Die Verteilung wird anhand der relavanten Attribute aus dem
	 * Gueltigkeitsbereich innerhalb des XML-Dokumentes initialisiert.
	 *
	 * @see XMLElement.
	 * 
	 * @param scope Der Gueltigkeitsbereich.
	 * @return true.
	 */
	bool init(const DOMElement* scope=NULL);
	
	//! Implementierung der Basisklasse TrafficDistribution.
	/*!
	 * Pakete koennen nur innerhalb von Aktivitaetsphasen generiert
	 * werden.
	 * 
	 * @param tick Der aktuelle Takt.
	 * @return true, falls laut Verteilung Paket generiert werden kann.
	 */
    virtual bool ifNowGenerate(unsigned int tick);

	//! Implementierung der Basisklasse TrafficDistribution.
	/*!
	 * Alle Klassenattribute, die den internen Zustand der Verteilung
	 * beschreiben, werden zurueckgesetzt.
	 */
	void reset();
	
	//! Destruktor.
	~ParetoDistribution(){};

protected:
	
	//! Zustandsuebergang.
	/*!
	 * Diese Methode veraendert den internen Zustand dieser Verteilung
	 * derart, dass eine Folge von Aktivitaets- und Ruhephasen entsteht.
	 */
	void advanceState();

	//! Bestimmung des Faktors fuer Groesse der Ruhephase.
	/*!
	 * Die Laenge der Ruhephasen (gaps) steht in Abhaengigkeit zu den 
	 * Formparametern fuer Aktivitaets- und Ruhephasen, sowie der 
	 * geforderten Last. Diese Methode bestimmt den aus diesen
	 * Abhaengigkeiten resultierenden Faktor.
	 *
	 * @return Faktor fuer fuer Groesse der Ruhephase.
	 */
	double calculateCoef() const;
	
	//! Umkehrungstransformation fuer Pareto-Verteilung.
	/*!
	 * Diese Methode bestimmt eine Zufallsvariable der Pareto-Verteilung
	 * anhand eines Formparemeters und der uniformern Verteilung (0,1).
	 * 
	 * @param shape Formparameter.
	 */
	double Xpareto(double shape) const;

private:

	// -- Benutzerangaben --
	
	double load;
	double onShape;		//shape parameter for burst distribution    
	double offShape;	
	
	// -- feste Parameter --

	unsigned int minBurst;
	double minGap;		// minimum gap value 
	
	// -- interner Zustand --

	unsigned int burstTick;
	unsigned int gapTick;
	unsigned int burstInterval;
	unsigned int gapInterval;
   	enum { GAP, BURST }state;
};

#endif
