#ifndef __PERIODIC_TRAFFIC__
#define __PERIODIC_TRAFFIC__

#include <traffic.h>

//! Periodischer Verkehr mit festem Zeitmuster.
/*!
 * Diese Klasse definiert eine Paketverteilung, die zu einem festen
 * Startzeitpunkt beginnt und sich dann periodisch fortsetzt.
 * 
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class PeriodicDistribution : public TrafficDistribution 
{
public:
    
	//! Konstruktor.
	/*!
	 * Der Konstruktor initialisiert lediglich die Attribute der Basisklassen
	 * und den Zaehler.
	 *
	 * @see TrafficDistribution
	 * @see XMLElement
	 * 
	 * @param psize Groesse der zu generierenden Pakete.
	 * @param scope Gueltigkeitsbereich innerhalb des XML-Dokumentes.
	 */
	PeriodicDistribution(unsigned int psize=1, const DOMElement* scope=NULL)
		: TrafficDistribution(psize,scope), tick(0) {}
	
	//! Implementierung der Basisklasse XMLElement.
	/*!
	 * Die Verteilung wird anhand der relavanten Attribute aus dem
	 * Gueltigkeitsbereich innerhalb des XML-Dokumentes initialisiert.
	 *
	 * @see XMLElement.
	 * 
	 * @param scope Der Gueltigkeitsbereich.
	 * @return true.
	 */
	bool init(const DOMElement* scope = NULL);

    //! Implementierung der Basisklasse TrafficDistribution.
	/*!
	 * Ein Paket kann generiert werden, falls der Starttakt erreicht wurde 
	 * und zu allen weiteren, durch den modulo festgelegten, Takten.
	 *
	 * @param tick Der aktuelle Takt.
	 * @return true, falls laut Verteilung Paket generiert werden kann.
	 */
	bool ifNowGenerate(unsigned int tick);

	//! Implementierung der Basisklasse TrafficDistribution.
	/*!
	 * Der interne Zustand dieser Verteilung wird durch den Taktzaehler
	 * bestimmt, welcher auf 0 zurueckgesetzt wird.
	 *
	 * FIXME: Der Taktzaehler ist nur ein Hilfsmittel, weil derzeit zwei
	 * Taktzaehler in der Simulation verwendet werden (obsTick und tick)
	 * und zwar in Abhaengigkeit von der Art der Simulation. Wenn das
	 * geklaert ist, sollte der uebergebene Systemtakt der Methode
	 * ifNowGenerate verwendet werden!
	 */
	void reset() { tick=0; }
	
	//! Destruktor.
	virtual ~PeriodicDistribution() {};
    
private:
	
	//! Taktzaehler
	unsigned int tick;
	//! Der Starttakt.
    unsigned int offset;
	//! Der Modulofaktor.
    unsigned int modulo;
};

#endif

