#include <randomburst.h>
#include <utility/xmlattribute.h>
#include <macros.h>
#include <environment/prng.h>

#include <cassert>

REGISTER_ELEMENT(RandomBurstsDistribution, "RandomBurst")

bool RandomBurstsDistribution::init(const DOMElement* scope)
{
	if(scope)
		setScope(scope);

	if(!Scope())
		_ERROR("Cannot initialize traffic distribution without its XML element!");

	BurstLoad = getAttribute("burstLoad");
	GapLoad = getAttribute("gapLoad");
	BurstIntervalMin = getAttribute("burstIntervalMin");
	BurstIntervalMax = getAttribute("burstIntervalMax");
	GapIntervalMin = getAttribute("gapIntervalMin");
	GapIntervalMax = getAttribute("gapIntervalMax");
	constPacketSize= getAttribute("packetSize", getParent());
	
	if(constPacketSize != 1) 
		_ERROR("[RandomBurst] Packet sizes larger than 1 are not supported yet.");
		
	if(BurstLoad < 0.0 || BurstLoad > 1.0)
		_ERROR("[RandomBurst] BurstLoad (" << BurstLoad << ") out of range!");
		
	if(GapLoad < 0.0 || GapLoad > 1.0)
		_ERROR("[RandomBurst] GapLoad (" << GapLoad << ") out of range!");
	
	if(BurstIntervalMin > BurstIntervalMax)
	{
		_WARNING("[RandomBurst] Bad interval (burst) [" 
				<< BurstIntervalMin << "," << BurstIntervalMax << "], will use ["
				<< BurstIntervalMax << "," << BurstIntervalMin << "], instead!");

		BurstIntervalMin = getAttribute("burstIntervalMax");
		BurstIntervalMax = getAttribute("burstIntervalMin");
	}
		
	if(GapIntervalMin > GapIntervalMax)
	{
		_WARNING("[RandomBurst] Bad interval (gap) [" 
				<< GapIntervalMin << "," << GapIntervalMax << "], will use ["
				<< GapIntervalMax << "," << GapIntervalMin << "], instead!");

		GapIntervalMin = getAttribute("gapIntervalMax");
		GapIntervalMax = getAttribute("gapIntervalMin");
	}

	// initialer Zustand
	reset();
	
	return true;
}

void RandomBurstsDistribution::reset()
{
	burstTick=0;
	gapTick=0;
	burstInterval=0;
	gapInterval=0;
	if( UniformDouble_1() > 0.5 ) state=BURST; else state=GAP;
}

bool RandomBurstsDistribution::ifNowGenerate(unsigned int)
{
	advanceState();
	
	// Burst
	if (state == BURST) 
	{
		return (BurstLoad>=UniformDouble_01());
	} 
	// Gap
	else
	{
		assert(state == GAP);
		
		return (GapLoad>=UniformDouble_01());
	}
}

void RandomBurstsDistribution::advanceState()
{
	if (state == BURST) 
	{ 
		// end of burst
		if (burstTick >= burstInterval) 
		{
			state=GAP;
			gapTick = 1;
			burstTick = 0;
			burstInterval = 0;
			
			gapInterval = (unsigned int) UniformInt(GapIntervalMin, GapIntervalMax);
		} 
		// still burst
		else burstTick++;
	} 
	
	else
	{
		assert(state == GAP);
		
		// end of gap
		if (gapTick >= gapInterval)
		{
			state=BURST;
			gapTick = 0;
			gapInterval = 0;
			burstTick = 1;
		
			burstInterval = (unsigned int) UniformInt(BurstIntervalMin, BurstIntervalMax);
		}
		// still gap
		else gapTick++;
	}
}
